/*-
 * SPDX-License-Identifier: BSD-2-Clause
 *
 * Copyright (c) 2016 Nicole Graziano <nicole@nextbsd.org>
 * All rights reserved.
 * Copyright (c) 2021 Rubicon Communications, LLC (Netgate)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

//RTX_PORT_START

#include "if_rtx.h"
#include "pcireg.h"
#include "igc_osdep.h"

#define if_ctx_t DRVDEV *

static void* iflib_get_softc(if_ctx_t ctx);

static int	igc_if_suspend(if_ctx_t ctx);
#ifndef UNDER_RTSS
static int	igc_if_tx_queues_alloc(if_ctx_t ctx, caddr_t* vaddrs, uint64_t* paddrs, int ntxqs, int ntxqsets);
static int	igc_if_rx_queues_alloc(if_ctx_t ctx, caddr_t* vaddrs, uint64_t* paddrs, int nrxqs, int nrxqsets);
static void	igc_if_queues_free(if_ctx_t ctx);
static uint64_t	igc_if_get_counter(if_ctx_t, ift_counter);
#endif

static void	igc_if_init(if_ctx_t ctx);
static void	igc_if_watchdog_reset(if_ctx_t ctx);

#ifndef UNDER_RTSS
static void	igc_if_media_status(if_ctx_t, struct ifmediareq*);
static int	igc_if_media_change(if_ctx_t ctx);
static int	igc_if_mtu_set(if_ctx_t ctx, uint32_t mtu);
static void	igc_if_vlan_register(if_ctx_t ctx, u16 vtag);
static void	igc_if_vlan_unregister(if_ctx_t ctx, u16 vtag);
static bool	igc_if_needs_restart(if_ctx_t ctx, enum iflib_restart_event event);
#endif

static void	igc_identify_hardware(if_ctx_t ctx);
static int	igc_allocate_pci_resources(if_ctx_t ctx);
static void	igc_free_pci_resources(if_ctx_t ctx);
static void	igc_reset(if_ctx_t ctx);
static int	igc_setup_msix(if_ctx_t ctx);

#ifndef UNDER_RTSS
static int	igc_setup_interface(if_ctx_t ctx);
#endif
static void	igc_initialize_transmit_unit(if_ctx_t ctx);
static void	igc_initialize_receive_unit(if_ctx_t ctx);


static void	igc_if_intr_enable(if_ctx_t ctx);
static void	igc_if_intr_disable(if_ctx_t ctx);
#ifndef UNDER_RTSS
static int	igc_if_rx_queue_intr_enable(if_ctx_t ctx, uint16_t rxqid);
static int	igc_if_tx_queue_intr_enable(if_ctx_t ctx, uint16_t txqid);
#endif

static void	igc_if_multi_set(if_ctx_t ctx, int flag);
#ifndef UNDER_RTSS
static void	igc_if_update_admin_status(if_ctx_t ctx);
static void	igc_if_debug(if_ctx_t ctx);
#endif
static void	igc_update_stats_counters(struct igc_adapter*);

static void	igc_add_hw_stats(struct igc_adapter* adapter);
static int	igc_if_set_promisc(if_ctx_t ctx, int rtflags);
static void	igc_setup_vlan_hw_support(struct igc_adapter*);
static int	igc_sysctl_nvm_info(SYSCTL_HANDLER_ARGS);
static void	igc_print_nvm_info(struct igc_adapter*);
static int	igc_sysctl_debug_info(SYSCTL_HANDLER_ARGS);
static int	igc_get_rs(SYSCTL_HANDLER_ARGS);
static void	igc_print_debug_info(struct igc_adapter*);
static int 	igc_is_valid_ether_addr(u8*);
static int	igc_sysctl_int_delay(SYSCTL_HANDLER_ARGS);
static void	igc_add_int_delay_sysctl(struct igc_adapter*, const char*,
	const char*, struct igc_int_delay_info*, int, int);
/* Management and WOL Support */
static void	igc_get_hw_control(struct igc_adapter*);
static void	igc_release_hw_control(struct igc_adapter*);
static void	igc_get_wakeup(if_ctx_t ctx);
static void	igc_enable_wakeup(if_ctx_t ctx);
#ifndef UNDER_RTSS
int		igc_intr(void* arg);

/* MSI-X handlers */
static int	igc_if_msix_intr_assign(if_ctx_t, int);
static int	igc_msix_link(void*);
static void	igc_handle_link(void* context);
static int	igc_set_flowcntl(SYSCTL_HANDLER_ARGS);
static int	igc_sysctl_eee(SYSCTL_HANDLER_ARGS);
#endif


//static void	igc_configure_queues(struct igc_adapter* adapter);
static void	igc_configure_queues(if_ctx_t ctx);

/*********************************************************************
 *  Tunable default values.
 *********************************************************************/

#define IGC_TICKS_TO_USECS(ticks)	((1024 * (ticks) + 500) / 1000)
#define IGC_USECS_TO_TICKS(usecs)	((1000 * (usecs) + 512) / 1024)

#define MAX_INTS_PER_SEC	8000
#define DEFAULT_ITR		(1000000000/(MAX_INTS_PER_SEC * 256))

static int igc_disable_crc_stripping = 0;

/*
** Tuneable Interrupt rate
*/
static int igc_max_interrupt_rate = 8000;
#ifndef UNDER_RTSS
SYSCTL_INT(_hw_igc, OID_AUTO, max_interrupt_rate, CTLFLAG_RDTUN,
	&igc_max_interrupt_rate, 0, "Maximum interrupts per second");

extern struct if_txrx igc_txrx;

static struct if_shared_ctx igc_sctx_init = {
	.isc_magic = IFLIB_MAGIC,
	.isc_q_align = PAGE_SIZE,
	.isc_tx_maxsize = IGC_TSO_SIZE + sizeof(struct ether_vlan_header),
	.isc_tx_maxsegsize = PAGE_SIZE,
	.isc_tso_maxsize = IGC_TSO_SIZE + sizeof(struct ether_vlan_header),
	.isc_tso_maxsegsize = IGC_TSO_SEG_SIZE,
	.isc_rx_maxsize = MAX_JUMBO_FRAME_SIZE,
	.isc_rx_nsegments = 1,
	.isc_rx_maxsegsize = MJUM9BYTES,
	.isc_nfl = 1,
	.isc_nrxqs = 1,
	.isc_ntxqs = 1,
	.isc_admin_intrcnt = 1,
	.isc_vendor_info = igc_vendor_info_array,
	.isc_driver_version = "1",
	.isc_driver = &igc_if_driver,
	.isc_flags = IFLIB_NEED_SCRATCH | IFLIB_TSO_INIT_IP | IFLIB_NEED_ZERO_CSUM,

	.isc_nrxd_min = {IGC_MIN_RXD},
	.isc_ntxd_min = {IGC_MIN_TXD},
	.isc_nrxd_max = {IGC_MAX_RXD},
	.isc_ntxd_max = {IGC_MAX_TXD},
	.isc_nrxd_default = {IGC_DEFAULT_RXD},
	.isc_ntxd_default = {IGC_DEFAULT_TXD},
};
#endif
/*****************************************************************
 *
 * Dump Registers
 *
 ****************************************************************/
#define IGC_REGS_LEN 739

int igc_get_regs(void* pCardInfo)
{
	struct igc_adapter* adapter = iflib_get_softc(pCardInfo);
	struct igc_hw* hw = &adapter->hw;
	DRVDEV* pCard = (DRVDEV*)pCardInfo;
	
	u32* regs_buff;
	int rc = 0;

	regs_buff = malloc(sizeof(u32) * IGC_REGS_LEN);
	memset(regs_buff, 0, IGC_REGS_LEN * sizeof(u32));

	/* General Registers */
	regs_buff[0] = IGC_READ_REG(hw, IGC_CTRL);
	regs_buff[1] = IGC_READ_REG(hw, IGC_STATUS);
	regs_buff[2] = IGC_READ_REG(hw, IGC_CTRL_EXT);
	regs_buff[3] = IGC_READ_REG(hw, IGC_ICR);
	regs_buff[4] = IGC_READ_REG(hw, IGC_RCTL);
	regs_buff[5] = IGC_READ_REG(hw, IGC_RDLEN(0));
	regs_buff[6] = IGC_READ_REG(hw, IGC_RDH(0));
	regs_buff[7] = IGC_READ_REG(hw, IGC_RDT(0));
	regs_buff[8] = IGC_READ_REG(hw, IGC_RXDCTL(0));
	regs_buff[9] = IGC_READ_REG(hw, IGC_RDBAL(0));
	regs_buff[10] = IGC_READ_REG(hw, IGC_RDBAH(0));
	regs_buff[11] = IGC_READ_REG(hw, IGC_TCTL);
	regs_buff[12] = IGC_READ_REG(hw, IGC_TDBAL(0));
	regs_buff[13] = IGC_READ_REG(hw, IGC_TDBAH(0));
	regs_buff[14] = IGC_READ_REG(hw, IGC_TDLEN(0));
	regs_buff[15] = IGC_READ_REG(hw, IGC_TDH(0));
	regs_buff[16] = IGC_READ_REG(hw, IGC_TDT(0));
	regs_buff[17] = IGC_READ_REG(hw, IGC_TXDCTL(0));

	RtPrintf("General Registers \n");
	RtPrintf("\tCTRL\t %08x\n", regs_buff[0]);
	RtPrintf("\tSTATUS\t %08x\n", regs_buff[1]);
	RtPrintf("\tCTRL_EXIT\t %08x\n\n", regs_buff[2]);

	RtPrintf("Interrupt Registers\n");
	RtPrintf("\tGPIE\t %08x\n", IGC_READ_REG(hw, IGC_GPIE));
	RtPrintf("\tEICR\t %08x\n", IGC_READ_REG(hw, IGC_EICR));
	RtPrintf("\tEICS\t %08x\n", IGC_READ_REG(hw, IGC_EICS));
	RtPrintf("\tEIMS\t %08x\n", IGC_READ_REG(hw, IGC_EIMS));
	RtPrintf("\tEIMC\t %08x\n", IGC_READ_REG(hw, IGC_EIMC));
	RtPrintf("\tEIAC\t %08x\n", IGC_READ_REG(hw, IGC_EIAC));
	RtPrintf("\tEIAM\t %08x\n", IGC_READ_REG(hw, IGC_EIAM));
	
	RtPrintf("\tICR\t %08x\n", regs_buff[3]);
	RtPrintf("\tICS\t %08x\n", IGC_READ_REG(hw, IGC_ICS));
	RtPrintf("\tIMS\t %08x\n", IGC_READ_REG(hw, IGC_IMS));
	RtPrintf("\tIMC\t %08x\n", IGC_READ_REG(hw, IGC_IMC));
	RtPrintf("\tIAM\t %08x\n\n", IGC_READ_REG(hw, IGC_IAM));	

	RtPrintf( "RX Registers\n");
	RtPrintf( "\tRCTL\t %08x\n", regs_buff[4]);
	RtPrintf( "\tRDLEN\t %08x\n", regs_buff[5]);
	RtPrintf( "\tRDH\t %08x\n", regs_buff[6]);
	RtPrintf( "\tRDT\t %08x\n", regs_buff[7]);
	RtPrintf( "\tRXDCTL\t %08x\n", regs_buff[8]);
	RtPrintf( "\tRDBAL\t %08x\n", regs_buff[9]);
	RtPrintf( "\tRDBAH\t %08x\n", regs_buff[10]);
	RtPrintf("\tIGC_MRQC\t %08x\n", IGC_READ_REG(hw, IGC_MRQC));
	RtPrintf("\tIGC_SRRCTL\t %08x\n\n", IGC_READ_REG(hw, IGC_SRRCTL(0)));

	RtPrintf( "TX Registers\n");
	RtPrintf( "\tTCTL\t %08x\n", regs_buff[11]);
	RtPrintf( "\tTDBAL\t %08x\n", regs_buff[12]);
	RtPrintf( "\tTDBAH\t %08x\n", regs_buff[13]);
	RtPrintf( "\tTDLEN\t %08x\n", regs_buff[14]);
	RtPrintf( "\tTDH\t %08x\n", regs_buff[15]);
	RtPrintf( "\tTDT\t %08x\n", regs_buff[16]);
	RtPrintf( "\tTXDCTL\t %08x\n", regs_buff[17]);
	RtPrintf( "\tTDFH\t %08x\n", regs_buff[18]);
	RtPrintf( "\tTDFT\t %08x\n", regs_buff[19]);
	RtPrintf( "\tTDFHS\t %08x\n", regs_buff[20]);
	RtPrintf( "\tTDFPC\t %08x\n\n", regs_buff[21]);

	free(regs_buff);

#ifdef DUMP_DESCS
	{
		int j;
		for (int QueNum = 0; QueNum < pCard->numberOfRxNicQueues; QueNum++)
		{
			DRV_QUE_DESCRIPTORS* pAdapterQ = &pCard->adapterQ[QueNum];
			for (j = 0; j < pAdapterQ->NumRxDescriptors; j++)
			{
				u32 staterr = le32toh(pAdapterQ->RxDescriptors[j].wb.upper.status_error);
				u32 length = le32toh(pAdapterQ->RxDescriptors[j].wb.upper.length);
				RtPrintf("Receive Desc(%d) Addr 0x%016I64x  pkt addr 0x%016I64x Error:%d  Length:%d\n", j, 
					&pAdapterQ->RxDescriptors[j],
					pAdapterQ->RxDescriptors[j].read.pkt_addr, staterr, length);
			}
		}
	}
#endif
	return(rc);
}

#define	IGC_CAPS							\
    IFCAP_HWCSUM | IFCAP_VLAN_MTU | IFCAP_VLAN_HWTAGGING |		\
    IFCAP_VLAN_HWCSUM | IFCAP_WOL | IFCAP_VLAN_HWFILTER | IFCAP_TSO4 |	\
    IFCAP_LRO | IFCAP_VLAN_HWTSO | IFCAP_JUMBO_MTU | IFCAP_HWCSUM_IPV6 |\
    IFCAP_TSO6

/*********************************************************************
 *  Device initialization routine
 *
 *  The attach entry point is called when the driver is being loaded.
 *  This routine identifies the type of hardware, allocates all resources
 *  and initializes the hardware.
 *
 *  return 0 on success, positive on failure
 *********************************************************************/


int
igc_if_attach_pre(void* pCardInfo)
{
	struct igc_adapter* adapter;
	device_t dev = (DRVDEV*)pCardInfo;
	DRVDEV* pCard = (DRVDEV*)pCardInfo;
	struct igc_hw* hw;
	int error = 0;

	INIT_DEBUGOUT("igc_if_attach_pre: begin");

	adapter = iflib_get_softc(pCardInfo);

	adapter->rx_num_queues = pCard->numberOfRxNicQueues;
	adapter->tx_num_queues = pCard->numberOfTxNicQueues;

	hw = &adapter->hw;

	/* Determine hardware and mac info */
	igc_identify_hardware(pCardInfo);
#ifndef UNDER_RTSS
	scctx->isc_tx_nsegments = IGC_MAX_SCATTER;
	scctx->isc_nrxqsets_max = scctx->isc_ntxqsets_max = igc_set_num_queues(ctx);
	if (bootverbose)
		device_printf(dev, "attach_pre capping queues at %d\n",
			scctx->isc_ntxqsets_max);

	scctx->isc_txqsizes[0] = roundup2(scctx->isc_ntxd[0] * sizeof(union igc_adv_tx_desc), IGC_DBA_ALIGN);
	scctx->isc_rxqsizes[0] = roundup2(scctx->isc_nrxd[0] * sizeof(union igc_adv_rx_desc), IGC_DBA_ALIGN);
	scctx->isc_txd_size[0] = sizeof(union igc_adv_tx_desc);
	scctx->isc_rxd_size[0] = sizeof(union igc_adv_rx_desc);
	scctx->isc_txrx = &igc_txrx;
	scctx->isc_tx_tso_segments_max = IGC_MAX_SCATTER;
	scctx->isc_tx_tso_size_max = IGC_TSO_SIZE;
	scctx->isc_tx_tso_segsize_max = IGC_TSO_SEG_SIZE;
	scctx->isc_capabilities = scctx->isc_capenable = IGC_CAPS;
	scctx->isc_tx_csum_flags = CSUM_TCP | CSUM_UDP | CSUM_TSO |
		CSUM_IP6_TCP | CSUM_IP6_UDP | CSUM_SCTP | CSUM_IP6_SCTP;

	/*
	** Some new devices, as with ixgbe, now may
	** use a different BAR, so we need to keep
	** track of which is used.
	*/
	scctx->isc_msix_bar = PCIR_BAR(IGC_MSIX_BAR);
	if (pci_read_config(dev, scctx->isc_msix_bar, 4) == 0)
		scctx->isc_msix_bar += 4;
#endif
	/* Setup PCI resources */
	if (igc_allocate_pci_resources(pCardInfo)) {
		device_printf(dev, "Allocation of PCI resources failed\n");
		error = ENXIO;
		goto err_pci;
	}

	/* Do Shared Code initialization */
	error = igc_setup_init_funcs(hw, true); 
	if (error) {
		device_printf(dev, "Setup of Shared code failed, error %d\n",
			error);
		error = ENXIO;
		goto err_pci;
	}
#ifndef UNDER_RTSS
	igc_setup_msix(ctx);
#endif

	igc_get_bus_info(hw);
#ifndef UNDER_RTSS
	/* Set up some sysctls for the tunable interrupt delays */
	igc_add_int_delay_sysctl(adapter, "rx_int_delay",
		"receive interrupt delay in usecs", &adapter->rx_int_delay,
		IGC_REGISTER(hw, IGC_RDTR), igc_rx_int_delay_dflt);
	igc_add_int_delay_sysctl(adapter, "tx_int_delay",
		"transmit interrupt delay in usecs", &adapter->tx_int_delay,
		IGC_REGISTER(hw, IGC_TIDV), igc_tx_int_delay_dflt);
	igc_add_int_delay_sysctl(adapter, "rx_abs_int_delay",
		"receive interrupt delay limit in usecs",
		&adapter->rx_abs_int_delay,
		IGC_REGISTER(hw, IGC_RADV),
		igc_rx_abs_int_delay_dflt);
	igc_add_int_delay_sysctl(adapter, "tx_abs_int_delay",
		"transmit interrupt delay limit in usecs",
		&adapter->tx_abs_int_delay,
		IGC_REGISTER(hw, IGC_TADV),
		igc_tx_abs_int_delay_dflt);
	igc_add_int_delay_sysctl(adapter, "itr",
		"interrupt delay limit in usecs/4",
		&adapter->tx_itr,
		IGC_REGISTER(hw, IGC_ITR),
		DEFAULT_ITR);
#endif

	hw->mac.autoneg = DO_AUTO_NEG;
	hw->phy.autoneg_wait_to_complete = false;
	hw->phy.autoneg_advertised = AUTONEG_ADV_DEFAULT;

	/* Copper options */
	if (hw->phy.media_type == igc_media_type_copper) {
		hw->phy.mdix = AUTO_ALL_MODES;
	}

#ifndef UNDER_RTSS
	/*
	 * Set the frame limits assuming
	 * standard ethernet sized frames.
	 */
	scctx->isc_max_frame_size = adapter->hw.mac.max_frame_size =
		ETHERMTU + ETHER_HDR_LEN + ETHERNET_FCS_SIZE;
#else
	adapter->hw.mac.max_frame_size = pCard->MaxPacket;
#endif
	/* Allocate multicast array memory. */
	adapter->mta = malloc(sizeof(u8) * ETHER_ADDR_LEN *
		MAX_NUM_MULTICAST_ADDRESSES);
	if (adapter->mta == NULL) {
		device_printf(dev, "Can not allocate multicast setup array\n");
		error = ENOMEM;
		goto err_late;
	}


	/* Check SOL/IDER usage */
	if (igc_check_reset_block(hw))
		device_printf(dev, "PHY reset is blocked"
			" due to SOL/IDER session.\n");
	
	adapter->hw.dev_spec._i225.eee_disable = 1;
#ifndef UNDER_RTSS
	/* Sysctl for setting Energy Efficient Ethernet */
	adapter->hw.dev_spec._i225.eee_disable = igc_eee_setting;
	SYSCTL_ADD_PROC(device_get_sysctl_ctx(dev),
		SYSCTL_CHILDREN(device_get_sysctl_tree(dev)),
		OID_AUTO, "eee_control",
		CTLTYPE_INT | CTLFLAG_RW | CTLFLAG_NEEDGIANT,
		adapter, 0, igc_sysctl_eee, "I",
		"Disable Energy Efficient Ethernet");
#endif
	/*
	** Start from a known state, this is
	** important in reading the nvm and
	** mac from that.
	*/
	igc_reset_hw(hw);

	/* Make sure we have a good EEPROM before we read from it */
	if (igc_validate_nvm_checksum(hw) < 0) {
		/*
		** Some PCI-E parts fail the first check due to
		** the link being in sleep state, call it again,
		** if it fails a second time its a real issue.
		*/
		if (igc_validate_nvm_checksum(hw) < 0) {
			device_printf(dev,
				"The EEPROM Checksum Is Not Valid\n");
			error = EIO;
			goto err_late;
		}
	}

	/* Copy the permanent MAC address out of the EEPROM */
	if (igc_read_mac_addr(hw) < 0) {
		device_printf(dev, "EEPROM read error while reading MAC"
			" address\n");
		error = EIO;
		goto err_late;
	}

	if (!igc_is_valid_ether_addr(hw->mac.addr)) {
		device_printf(dev, "Invalid MAC address\n");
		error = EIO;
		goto err_late;
	}

	/*
	 * Get Wake-on-Lan and Management info for later use
	 */
	igc_get_wakeup(pCardInfo);
#ifndef UNDER_RTSS
	/* Enable only WOL MAGIC by default */
	scctx->isc_capenable &= ~IFCAP_WOL;
	if (adapter->wol != 0)
		scctx->isc_capenable |= IFCAP_WOL_MAGIC;

	iflib_set_mac(ctx, hw->mac.addr);
#endif
	INIT_DEBUGOUT("igc_if_attach_pre: end");
	return (0);

err_late:
	igc_release_hw_control(adapter);
err_pci:
	igc_free_pci_resources(pCardInfo);
	free(adapter->mta);

	return (error);
}

int
igc_if_attach_post(void* pCardInfo)
{
	struct igc_adapter* adapter = iflib_get_softc(pCardInfo);
	struct igc_hw* hw = &adapter->hw;
	int error = 0;
	INIT_DEBUGOUT("igc_if_attach_post: begin");

	igc_reset(pCardInfo);

	/* Initialize statistics */
	igc_update_stats_counters(adapter);
	hw->mac.get_link_status = true;
#ifndef UNDER_RTSS
	igc_if_update_admin_status(ctx);
#endif
	igc_add_hw_stats(adapter);

	/* the driver can now take control from firmware */
	igc_get_hw_control(adapter);

	INIT_DEBUGOUT("igc_if_attach_post: end");

	return (error);
}


/*********************************************************************
 *  Device removal routine
 *
 *  The detach entry point is called when the driver is being removed.
 *  This routine stops the adapter and deallocates all the resources
 *  that were allocated for driver operation.
 *
 *  return 0 on success, positive on failure
 *********************************************************************/
int
igc_if_detach(void *pCardInfo)
{
	struct igc_adapter* adapter = iflib_get_softc(pCardInfo);

	INIT_DEBUGOUT("igc_if_detach: begin");

	igc_phy_hw_reset(&adapter->hw);

	igc_release_hw_control(adapter);
//RTX Port
	if (adapter->mta != NULL)
		free(adapter->mta);
//RTX Port End
	igc_free_pci_resources(pCardInfo);

	return (0);
}

/*********************************************************************
 *
 *  Shutdown entry point
 *
 **********************************************************************/

int
igc_if_shutdown(void *pCardInfo)
{
	return igc_if_suspend(pCardInfo);
}

/*
 * Suspend/resume device methods.
 */
static int
igc_if_suspend(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);

	igc_release_hw_control(adapter);
	igc_enable_wakeup(ctx);
	return (0);
}

#ifndef UNDER_RTSS

static int
igc_if_mtu_set(if_ctx_t ctx, uint32_t mtu)
{
	int max_frame_size;
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	if_softc_ctx_t scctx = iflib_get_softc_ctx(ctx);

	IOCTL_DEBUGOUT("ioctl rcv'd: SIOCSIFMTU (Set Interface MTU)");

	/* 9K Jumbo Frame size */
	max_frame_size = 9234;

	if (mtu > max_frame_size - ETHER_HDR_LEN - ETHER_CRC_LEN) {
		return (EINVAL);
	}

	scctx->isc_max_frame_size = adapter->hw.mac.max_frame_size =
		mtu + ETHER_HDR_LEN + ETHER_CRC_LEN;
	return (0);
}
#endif

/*********************************************************************
 *  Init entry point
 *
 *  This routine is used in two ways. It is used by the stack as
 *  init entry point in network interface structure. It is also used
 *  by the driver as a hw/sw initialization routine to get to a
 *  consistent state.
 *
 **********************************************************************/

static void
igc_if_init(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);

	//struct igc_tx_queue* tx_que;
	DRVDEV* pCardInfo = (DRVDEV*)ctx;

	INIT_DEBUGOUT("igc_if_init: begin");
#ifndef UNDER_RTSS
	/* Get the latest mac address, User can use a LAA */
	bcopy(if_getlladdr(ifp), adapter->hw.mac.addr,
		ETHER_ADDR_LEN);
#endif
	/* Put the address into the Receive Address Array */
	igc_rar_set(&adapter->hw, adapter->hw.mac.addr, 0);

	/* Initialize the hardware */
	igc_reset(ctx);

#ifndef UNDER_RTSS
	igc_if_update_admin_status(ctx);

	for (i = 0, tx_que = adapter->tx_queues; i < adapter->tx_num_queues; i++, tx_que++) {
		struct tx_ring* txr = &tx_que->txr;

		txr->tx_rs_cidx = txr->tx_rs_pidx;

		/* Initialize the last processed descriptor to be the end of
		 * the ring, rather than the start, so that we avoid an
		 * off-by-one error when calculating how many descriptors are
		 * done in the credits_update function.
		 */
		txr->tx_cidx_processed = scctx->isc_ntxd[0] - 1;
	}

#endif
	/* Setup VLAN support, basic and offload if available */
	IGC_WRITE_REG(&adapter->hw, IGC_VET, ETHERTYPE_VLAN);

	/* Prepare transmit descriptors and buffers */
	igc_initialize_transmit_unit(ctx);

	/* Setup Multicast table */
	igc_if_multi_set(ctx, IFF_ALLMULTI);

	adapter->rx_mbuf_sz = pCardInfo->RxBufferSize;// iflib_get_rx_mbuf_sz(ctx);
	igc_initialize_receive_unit(ctx);
#ifndef UNDER_RTSS

	/* Use real VLAN Filter support? */
	if (if_getcapenable(ifp) & IFCAP_VLAN_HWTAGGING) {
		if (if_getcapenable(ifp) & IFCAP_VLAN_HWFILTER)
			/* Use real VLAN Filter support */
			igc_setup_vlan_hw_support(adapter);
		else {
			u32 ctrl;
			ctrl = IGC_READ_REG(&adapter->hw, IGC_CTRL);
			ctrl |= IGC_CTRL_VME;
			IGC_WRITE_REG(&adapter->hw, IGC_CTRL, ctrl);
		}
	}
#endif

	/* Don't lose promiscuous settings */
	//igc_if_set_promisc(ctx, ENIOCPROMISC);
	igc_clear_hw_cntrs_base_generic(&adapter->hw);


	if (adapter->intr_type == IFLIB_INTR_MSIX) /* Set up queue routing */
		igc_configure_queues(pCardInfo);
	else
	{
		IGC_WRITE_REG(&adapter->hw, IGC_GPIE,
			IGC_GPIE_NSICR);

		IGC_WRITE_FLUSH(&adapter->hw);

		IGC_WRITE_REG(&adapter->hw, IGC_EITR(0), 0);
		IGC_WRITE_FLUSH(&adapter->hw);
	}
	/* this clears any pending interrupts */
	IGC_READ_REG(&adapter->hw, IGC_ICR);

	/* the driver can now take control from firmware */
	igc_get_hw_control(adapter);

	/* Set Energy Efficient Ethernet */
	igc_set_eee_i225(&adapter->hw, true, true, true);

}

#ifndef UNDER_RTSS
/*********************************************************************
 *
 *  Fast Legacy/MSI Combined Interrupt Service routine
 *
 *********************************************************************/
int
igc_intr(void* arg)
{
	struct igc_adapter* adapter = arg;
	if_ctx_t ctx = adapter->ctx;
	u32 reg_icr;

	reg_icr = IGC_READ_REG(&adapter->hw, IGC_ICR);

	/* Hot eject? */
	if (reg_icr == 0xffffffff)
		return FILTER_STRAY;

	/* Definitely not our interrupt. */
	if (reg_icr == 0x0)
		return FILTER_STRAY;

	if ((reg_icr & IGC_ICR_INT_ASSERTED) == 0)
		return FILTER_STRAY;

	/*
	 * Only MSI-X interrupts have one-shot behavior by taking advantage
	 * of the EIAC register.  Thus, explicitly disable interrupts.  This
	 * also works around the MSI message reordering errata on certain
	 * systems.
	 */
	IFDI_INTR_DISABLE(ctx);

	/* Link status change */
	if (reg_icr & (IGC_ICR_RXSEQ | IGC_ICR_LSC))
		igc_handle_link(ctx);

	if (reg_icr & IGC_ICR_RXO)
		adapter->rx_overruns++;

	return (FILTER_SCHEDULE_THREAD);
}

static int
igc_if_rx_queue_intr_enable(if_ctx_t ctx, uint16_t rxqid)
{
#if 0 //ToDo
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct igc_rx_queue* rxq = &adapter->rx_queues[rxqid];

	IGC_WRITE_REG(&adapter->hw, IGC_EIMS, rxq->eims);
#endif
	return (0);
}

static int
igc_if_tx_queue_intr_enable(if_ctx_t ctx, uint16_t txqid)
{
#if 0 //ToDo
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct igc_tx_queue* txq = &adapter->tx_queues[txqid];

	IGC_WRITE_REG(&adapter->hw, IGC_EIMS, txq->eims);
#endif
	return (0);
}

/*********************************************************************
 *
 *  MSI-X RX Interrupt Service routine
 *
 **********************************************************************/
static int
igc_msix_que(void* arg)
{

	struct igc_rx_queue* que = arg;

	++que->irqs;

	return (FILTER_SCHEDULE_THREAD);
}

/*********************************************************************
 *
 *  MSI-X Link Fast Interrupt Service routine
 *
 **********************************************************************/
static int
igc_msix_link(void* arg)
{
	struct igc_adapter* adapter = arg;
	u32 reg_icr;

	++adapter->link_irq;
	MPASS(adapter->hw.back != NULL);
	reg_icr = IGC_READ_REG(&adapter->hw, IGC_ICR);

	if (reg_icr & IGC_ICR_RXO)
		adapter->rx_overruns++;

	if (reg_icr & (IGC_ICR_RXSEQ | IGC_ICR_LSC)) {
		igc_handle_link(adapter->ctx);
	}

	IGC_WRITE_REG(&adapter->hw, IGC_IMS, IGC_IMS_LSC);
	IGC_WRITE_REG(&adapter->hw, IGC_EIMS, adapter->link_mask);

	return (FILTER_HANDLED);
}

static void
igc_handle_link(void* context)
{
	if_ctx_t ctx = context;
	struct igc_adapter* adapter = iflib_get_softc(ctx);

	adapter->hw.mac.get_link_status = true;
	iflib_admin_intr_deferred(ctx);
}

/*********************************************************************
 *
 *  Media Ioctl callback
 *
 *  This routine is called whenever the user queries the status of
 *  the interface using ifconfig.
 *
 **********************************************************************/
static void
igc_if_media_status(if_ctx_t ctx, struct ifmediareq* ifmr)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);

	INIT_DEBUGOUT("igc_if_media_status: begin");

	iflib_admin_intr_deferred(ctx);

	ifmr->ifm_status = IFM_AVALID;
	ifmr->ifm_active = IFM_ETHER;

	if (!adapter->link_active) {
		return;
	}

	ifmr->ifm_status |= IFM_ACTIVE;

	switch (adapter->link_speed) {
	case 10:
		ifmr->ifm_active |= IFM_10_T;
		break;
	case 100:
		ifmr->ifm_active |= IFM_100_TX;
		break;
	case 1000:
		ifmr->ifm_active |= IFM_1000_T;
		break;
	case 2500:
		ifmr->ifm_active |= IFM_2500_T;
		break;
	}

	if (adapter->link_duplex == FULL_DUPLEX)
		ifmr->ifm_active |= IFM_FDX;
	else
		ifmr->ifm_active |= IFM_HDX;
}

/*********************************************************************
 *
 *  Media Ioctl callback
 *
 *  This routine is called when the user changes speed/duplex using
 *  media/mediopt option with ifconfig.
 *
 **********************************************************************/
static int
igc_if_media_change(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct ifmedia* ifm = iflib_get_media(ctx);

	INIT_DEBUGOUT("igc_if_media_change: begin");
#if 0 //ToDo
	if (IFM_TYPE(ifm->ifm_media) != IFM_ETHER)
		return (EINVAL);

	adapter->hw.mac.autoneg = DO_AUTO_NEG;

	switch (IFM_SUBTYPE(ifm->ifm_media)) {
	case IFM_AUTO:
		adapter->hw.phy.autoneg_advertised = AUTONEG_ADV_DEFAULT;
		break;
	case IFM_2500_T:
		adapter->hw.phy.autoneg_advertised = ADVERTISE_2500_FULL;
		break;
	case IFM_1000_T:
		adapter->hw.phy.autoneg_advertised = ADVERTISE_1000_FULL;
		break;
	case IFM_100_TX:
		if ((ifm->ifm_media & IFM_GMASK) == IFM_HDX)
			adapter->hw.phy.autoneg_advertised = ADVERTISE_100_HALF;
		else
			adapter->hw.phy.autoneg_advertised = ADVERTISE_100_FULL;
		break;
	case IFM_10_T:
		if ((ifm->ifm_media & IFM_GMASK) == IFM_HDX)
			adapter->hw.phy.autoneg_advertised = ADVERTISE_10_HALF;
		else
			adapter->hw.phy.autoneg_advertised = ADVERTISE_10_FULL;
		break;
	default:
		device_printf(adapter->dev, "Unsupported media type\n");
	}
#endif 

	igc_if_init(ctx);

	return (0);
}
#endif

static int
igc_if_set_promisc(if_ctx_t ctx, int rtflags)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);

	u32 reg_rctl;
	int mcnt = 0;
	int flags;
	if (rtflags == ENIOCPROMISC || rtflags == ENIOCALL)
	{
		flags = IFF_PROMISC;
	}

	reg_rctl = IGC_READ_REG(&adapter->hw, IGC_RCTL);
	reg_rctl &= ~(IGC_RCTL_SBP | IGC_RCTL_UPE);
	if (flags & IFF_ALLMULTI)
		mcnt = MAX_NUM_MULTICAST_ADDRESSES;
	else
		mcnt = 128;//RTX Port min(if_llmaddr_count(ifp), MAX_NUM_MULTICAST_ADDRESSES);

	/* Don't disable if in MAX groups */
	if (mcnt < MAX_NUM_MULTICAST_ADDRESSES)
		reg_rctl &= (~IGC_RCTL_MPE);
	IGC_WRITE_REG(&adapter->hw, IGC_RCTL, reg_rctl);

	if (flags & IFF_PROMISC) {
		reg_rctl |= (IGC_RCTL_UPE | IGC_RCTL_MPE);
		/* Turn this on if you want to see bad packets */
		//if (igc_debug_sbp) RTX PORT
		if (rtflags == ENIOCALL)
			reg_rctl |= IGC_RCTL_SBP;

		IGC_WRITE_REG(&adapter->hw, IGC_RCTL, reg_rctl);
	}
	else if (flags & IFF_ALLMULTI) {
		reg_rctl |= IGC_RCTL_MPE;
		reg_rctl &= ~IGC_RCTL_UPE;
		IGC_WRITE_REG(&adapter->hw, IGC_RCTL, reg_rctl);
	}

	return (0);
}
#ifndef UNDER_RTSS
static u_int
igc_copy_maddr(void* arg, struct sockaddr_dl* sdl, u_int idx)
{
	u8* mta = arg;

	if (idx == MAX_NUM_MULTICAST_ADDRESSES)
		return (0);

	bcopy(LLADDR(sdl), &mta[idx * ETHER_ADDR_LEN], ETHER_ADDR_LEN);

	return (1);
}
#endif

/*********************************************************************
 *  Multicast Update
 *
 *  This routine is called whenever multicast address list is updated.
 *
 **********************************************************************/

static void
igc_if_multi_set(if_ctx_t ctx, int flag)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);

	u8* mta; /* Multicast array memory */
	u32 reg_rctl = 0;
	int mcnt = 0;

	IOCTL_DEBUGOUT("igc_set_multi: begin");

	mta = adapter->mta;
	//bzero(mta, sizeof(u8) * ETHER_ADDR_LEN * MAX_NUM_MULTICAST_ADDRESSES);

	//mcnt = if_foreach_llmaddr(ifp, igc_copy_maddr, mta);
	mcnt = 128;//RTX Port;

	reg_rctl = IGC_READ_REG(&adapter->hw, IGC_RCTL);

	if (flag == IFF_PROMISC) {
		reg_rctl |= (IGC_RCTL_UPE | IGC_RCTL_MPE);
		/* Turn this on if you want to see bad packets */
		//if (igc_debug_sbp)
		//	reg_rctl |= IGC_RCTL_SBP;
	}
	else if (mcnt >= MAX_NUM_MULTICAST_ADDRESSES ||
		(flag == IFF_ALLMULTI)) {
		reg_rctl |= IGC_RCTL_MPE;
		reg_rctl &= ~IGC_RCTL_UPE;
	}
	else
		reg_rctl &= ~(IGC_RCTL_UPE | IGC_RCTL_MPE);

	if (mcnt < MAX_NUM_MULTICAST_ADDRESSES)
		igc_update_mc_addr_list(&adapter->hw, mta, mcnt);

	IGC_WRITE_REG(&adapter->hw, IGC_RCTL, reg_rctl);
}


static void
igc_if_update_admin_status(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct igc_hw* hw = &adapter->hw;
	device_t dev = ctx;
	DRVDEV* pCardInfo = ctx;
	u32 link_check, thstat, ctrl;

	link_check = thstat = ctrl = 0;
	/* Get the cached link value or read phy for real */
	switch (hw->phy.media_type) {
	case igc_media_type_copper:
		if (hw->mac.get_link_status == true) {
			/* Do the work to read phy */
			igc_check_for_link(hw);
			link_check = !hw->mac.get_link_status;
		}
		else
			link_check = true;
		break;
	case igc_media_type_unknown:
		igc_check_for_link(hw);
		link_check = !hw->mac.get_link_status;
		/* FALLTHROUGH */
	default:
		break;
	}

	/* Now check for a transition */
	if (link_check && (adapter->link_active == 0)) {
		igc_get_speed_and_duplex(hw, &adapter->link_speed,
			&adapter->link_duplex);
		//if (bootverbose)
		//	device_printf(dev, "Link is up %d Mbps %s\n",
		//		adapter->link_speed,
		//		((adapter->link_duplex == FULL_DUPLEX) ?
		//			"Full Duplex" : "Half Duplex"));
		adapter->link_active = 1;
		//iflib_link_state_change(ctx, LINK_STATE_UP,
		//	IF_Mbps(adapter->link_speed));
		pCardInfo->LinkIsActive = RtndMediaConnectStateConnected;
		pCardInfo->LineSpeed = adapter->link_speed;
		pCardInfo->FullDuplex = adapter->link_duplex;
	}
	else if (!link_check && (adapter->link_active == 1)) {
		adapter->link_speed = 0;
		adapter->link_duplex = 0;
		adapter->link_active = 0;
		//iflib_link_state_change(ctx, LINK_STATE_DOWN, 0);
		pCardInfo->LinkIsActive = RtndMediaConnectStateDisConnected;
		pCardInfo->LineSpeed = adapter->link_speed;
		pCardInfo->FullDuplex = adapter->link_duplex;
	}
	igc_update_stats_counters(adapter);
}

static void
igc_if_watchdog_reset(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);

	/*
	 * Just count the event; iflib(4) will already trigger a
	 * sufficient reset of the controller.
	 */
	adapter->watchdog_events++;
}

/*********************************************************************
 *
 *  This routine disables all traffic on the adapter by issuing a
 *  global reset on the MAC.
 *
 **********************************************************************/
void
igc_if_stop(void * ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);

	INIT_DEBUGOUT("igc_if_stop: begin");

	igc_reset_hw(&adapter->hw);
	IGC_WRITE_REG(&adapter->hw, IGC_WUC, 0);
	INIT_DEBUGOUT("igc_if_stop: end");
	
}

/*********************************************************************
 *
 *  Determine hardware revision.
 *
 **********************************************************************/
static void
igc_identify_hardware(if_ctx_t ctx)
{

	//device_t dev = iflib_get_dev(ctx);
	device_t dev = ctx;

	struct igc_adapter* adapter = iflib_get_softc(ctx);

	/* Make sure our PCI config space has the necessary stuff set */
	adapter->hw.bus.pci_cmd_word = pci_read_config(dev, PCIR_COMMAND, 2);

	/* Save off the information about this board */
	adapter->hw.vendor_id = ctx->VendorId;
	adapter->hw.device_id = ctx->DeviceId;
	adapter->hw.revision_id = pci_read_config(dev, PCIR_REVID, 1);
	adapter->hw.subsystem_vendor_id =
		pci_read_config(dev, PCIR_SUBVEND_0, 2);
	adapter->hw.subsystem_device_id =
		pci_read_config(dev, PCIR_SUBDEV_0, 2);

	/* Do Shared Code Init and Setup */
	if (igc_set_mac_type(&adapter->hw)) {
		device_printf(dev, "Setup init failure\n");
		return;
	}

}


static int
igc_allocate_pci_resources(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	device_t dev = ctx;
	DRVDEV* pCardInfo = ctx;
	int rid;

	rid = PCIR_BAR(0);
	adapter->memory = bus_alloc_resource_any(dev, SYS_RES_MEMORY,
		&rid, RF_ACTIVE);
	if (adapter->memory == NULL) {
		device_printf(dev, "Unable to allocate bus resource: memory\n");
		return (ENXIO);
	}
	adapter->osdep.mem_bus_space_tag = (bus_space_tag_t)NULL;  // rman_get_bustag(adapter->memory);
	adapter->osdep.mem_bus_space_handle = (bus_space_handle_t)adapter->flash; //	rman_get_bushandle(adapter->memory);
	//adapter->hw.hw_addr = (u8*)&adapter->osdep.mem_bus_space_handle;
	//adapter->hw.back = &adapter->osdep;
//RTX Port Start
	adapter->hw.hw_addr = pCardInfo->CsrBase;
	adapter->hw.back = pCardInfo;
//RTX Port End
	return (0);
}
#ifndef UNDER_RTSS
/*********************************************************************
 *
 *  Set up the MSI-X Interrupt handlers
 *
 **********************************************************************/
static int
igc_if_msix_intr_assign(if_ctx_t ctx, int msix)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
#if 0 //ToDo
	struct igc_rx_queue* rx_que = adapter->rx_queues;
	struct igc_tx_queue* tx_que = adapter->tx_queues;
	int error, rid, i, vector = 0, rx_vectors;
	char buf[16];

	/* First set up ring resources */
	for (i = 0; i < adapter->rx_num_queues; i++, rx_que++, vector++) {
		rid = vector + 1;
		snprintf(buf, sizeof(buf), "rxq%d", i);
		error = iflib_irq_alloc_generic(ctx, &rx_que->que_irq, rid, IFLIB_INTR_RXTX, igc_msix_que, rx_que, rx_que->me, buf);
		if (error) {
			device_printf(iflib_get_dev(ctx), "Failed to allocate que int %d err: %d", i, error);
			adapter->rx_num_queues = i + 1;
			goto fail;
		}

		rx_que->msix = vector;

		/*
		 * Set the bit to enable interrupt
		 * in IGC_IMS -- bits 20 and 21
		 * are for RX0 and RX1, note this has
		 * NOTHING to do with the MSI-X vector
		 */
		rx_que->eims = 1 << vector;
	}
	rx_vectors = vector;

	vector = 0;
	for (i = 0; i < adapter->tx_num_queues; i++, tx_que++, vector++) {
		snprintf(buf, sizeof(buf), "txq%d", i);
		tx_que = &adapter->tx_queues[i];
		iflib_softirq_alloc_generic(ctx,
			&adapter->rx_queues[i % adapter->rx_num_queues].que_irq,
			IFLIB_INTR_TX, tx_que, tx_que->me, buf);

		tx_que->msix = (vector % adapter->rx_num_queues);

		/*
		 * Set the bit to enable interrupt
		 * in IGC_IMS -- bits 22 and 23
		 * are for TX0 and TX1, note this has
		 * NOTHING to do with the MSI-X vector
		 */
		tx_que->eims = 1 << i;
	}

	/* Link interrupt */
	rid = rx_vectors + 1;
	error = iflib_irq_alloc_generic(ctx, &adapter->irq, rid, IFLIB_INTR_ADMIN, igc_msix_link, adapter, 0, "aq");

	if (error) {
		device_printf(iflib_get_dev(ctx), "Failed to register admin handler");
		goto fail;
	}
	adapter->linkvec = rx_vectors;
	return (0);
fail:
	iflib_irq_free(ctx, &adapter->irq);
	rx_que = adapter->rx_queues;
	for (int i = 0; i < adapter->rx_num_queues; i++, rx_que++)
		iflib_irq_free(ctx, &rx_que->que_irq);

	return (error);
#endif
	return 0;
}
#endif
static void
igc_configure_queues(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct igc_hw* hw = &adapter->hw;
	DRVDEV* pCardInfo = ctx;
#ifndef UNDER_RTSS
	struct igc_rx_queue* rx_que;
	struct igc_tx_queue* tx_que;
#endif
	u32 ivar = 0, newitr = 0;

	ULONG iVarValue = 0;
	ULONG numberOfQueues = 0;
	ULONG rxMessageCount = 0;
	ULONG txMessageCount = 0;
	ULONG messageCount = 0;
	ULONG queue = 0;
	ULONG rxMessage = 0;
	ULONG rxOffset = 0;
	ULONG txMessage = 0;
	ULONG i = 0;
	ULONG           TotalMessagesInCapability;
	BYTE            FreeMessageIdBitMap[4];
	ULONG           FreeMessages, MessageCount;
	ULONG           ProcessorFreeVectors[MAXIMUM_PROCESSORS] = { 0 };

	// Query device MSIX free messages and RTSS system free vectors
	FreeMessages = RtGetPciMsixFreeMessages(pCardInfo->BusNumber, pCardInfo->SlotNumber,
		&TotalMessagesInCapability, FreeMessageIdBitMap);

	// Requred message count is driven by the number of Rx and Tx queues for an interface
		// Plus one more for line status
	MessageCount = pCardInfo->numberOfRxNicQueues +
		pCardInfo->numberOfTxNicQueues + 1;

	// Error out if there are not enough MSIX messages we will be reusing them
	// so set to use the max amount
	if (FreeMessages < MessageCount)
	{
		MessageCount = FreeMessages;
	}

	// Keep for debugging
		//RtPrintf("HWINIT: MESSAGES -> Free: %d. Ideal: %d Used: %d \n", FreeMessages, MessageCount, MessageCount/2);

	MessageCount /= 2;

	pCardInfo->usedMessagesCount = MessageCount;

	ULONG txOffset = 0;

	// The maximum number of queues for fill IVAR for
	numberOfQueues = (pCardInfo->numberOfRxNicQueues > pCardInfo->numberOfTxNicQueues) ?
		pCardInfo->numberOfRxNicQueues : pCardInfo->numberOfTxNicQueues;

	/* First turn on RSS capability */
	IGC_WRITE_REG(hw, IGC_GPIE, IGC_GPIE_MSIX_MODE |
		IGC_GPIE_PBA| IGC_GPIE_EIAME | 
		IGC_GPIE_NSICR);

	IGC_WRITE_FLUSH(hw);

	// Setup IVAR for all the queues.
// For each queue index there could be up to two messages (Rx/Tx) 
	for (queue = 0; queue < numberOfQueues; queue++)
	{
		// Keep for debugging
		//RtPrintf("I225 HardwareInit QUEUE %d. MESSAGES: %d  \n" , 
		//	queue ,
		//	pCardInfo->usedMessagesCount);

		// Next message. NOTE: First one is - 1
		messageCount++;

		// Define IVAR and message-to-queue map . 
		// If there are more Rx queues to fill - fill Rx part 
		if (queue < (ULONG)pCardInfo->numberOfRxNicQueues)
		{
			// Skip message 0 as it's used for line status
			if (0 == (messageCount % pCardInfo->usedMessagesCount))
			{
				messageCount++;
			}

			// Reuse messages when ran out of free ones. Maximum amount is limited by the PCI device.  
			// Message ID cannot be less than 1 
			rxMessage = (1 == pCardInfo->usedMessagesCount) ? 1 : (messageCount % pCardInfo->usedMessagesCount);

			// Even queues - Rx offset: 0. Odd queues Rx offset: 16.
			rxOffset = 16 * (rxMessageCount % 2);
			// Update the current IVAR value to-be-set for Rx
			iVarValue |= (0x80 | rxMessage) << rxOffset;
			rxMessageCount++;

			// Fill message-to-queue map
			pCardInfo->messageToInerruptKindQueueMap[rxMessage].messageId = rxMessage;
			pCardInfo->messageToInerruptKindQueueMap[rxMessage].queues[pCardInfo->messageToInerruptKindQueueMap[rxMessage].numberOfQueues].kind = InterruptKindReceive;
			pCardInfo->messageToInerruptKindQueueMap[rxMessage].queues[pCardInfo->messageToInerruptKindQueueMap[rxMessage].numberOfQueues].queue = queue;
			pCardInfo->messageToInerruptKindQueueMap[rxMessage].numberOfQueues++;

			// Keep for debugging
			//RtPrintf("RX IVAR[%d] = 0x%x. Q: %d, MSG: %d. RX MSG COUNT: %d , OFFSET: %d \n", queue / 2, iVarValue, queue, rxMessage , rxMessageCount, rxOffset);

		}

		// Next message
		messageCount++;

		// If there are more Tx queues to fill - fill Tx part 
		if (queue < (ULONG)pCardInfo->numberOfTxNicQueues)
		{

			// Skip message 0 as it's used for line status
			if (0 == (messageCount % pCardInfo->usedMessagesCount))
			{
				messageCount++;
			}

			// Reuse messages when ran out of free ones. Maximum amount is limited by the PCI device.
			// Message ID cannot be less than 1 
			txMessage = (1 == pCardInfo->usedMessagesCount) ? 1 : (messageCount % pCardInfo->usedMessagesCount);

			// Even queues - Tx offset: 8. Odd queues Tx offset: 24.
			txOffset = 8 + 16 * (txMessageCount % 2);
			// Update the current IVAR value to-be-set for Tx
			iVarValue |= (0x80 | txMessage) << txOffset;
			txMessageCount++;

			// Fill message-to-queue map
			pCardInfo->messageToInerruptKindQueueMap[txMessage].messageId = txMessage;
			pCardInfo->messageToInerruptKindQueueMap[txMessage].queues[pCardInfo->messageToInerruptKindQueueMap[txMessage].numberOfQueues].kind = InterruptKindTransmit;
			pCardInfo->messageToInerruptKindQueueMap[txMessage].queues[pCardInfo->messageToInerruptKindQueueMap[txMessage].numberOfQueues].queue = queue;
			pCardInfo->messageToInerruptKindQueueMap[txMessage].numberOfQueues++;

			// Keep for debugging
			//RtPrintf("TX IVAR[%d] = 0x%x. Q: %d, MSG: %d. TX MSG COUNT: %d, OFFSET: %d \n", queue / 2, iVarValue, queue, txMessage , txMessageCount, txOffset);

		}

		// For each two queues, update IVAR register. 
		// If there are less than 2 queues - update the register right away
		if ((numberOfQueues < 2) || ((queue + 1) % 2) == 0)
		{
			// Keep for debugging
			//RtPrintf("SET IVAR[%d]{0x%x} = 0x%x \n",
			//	queue / 2, iVarRegisterAssignement[queue / 2], iVarValue);

			// Write the current IVAR value
			//REG_WRITE(iVarRegisterAssignement[queue / 2], iVarValue);
			IGC_WRITE_REG_ARRAY(hw, IGC_IVAR0, queue / 2, iVarValue);
			IGC_WRITE_FLUSH(hw);
			// Reset iVarValue 
			iVarValue = 0;
		};
	}
	// [Timer interrupt]: MessageId = 0; [Other Cause interrupts]: MessageId = 0
	IGC_WRITE_REG(hw, IGC_IVAR_MISC, (0x80 | 0x00) + ((0x80 | 0x00) << 8));
	IGC_WRITE_FLUSH(hw);

	/* Set the starting interrupt rate */
	if (igc_max_interrupt_rate > 0)
		newitr = (4000000 / igc_max_interrupt_rate) & 0x7FFC;

	newitr |= IGC_EITR_CNT_IGNR;

	for (int i = 0; i < adapter->rx_num_queues; i++) {
		IGC_WRITE_REG(hw, IGC_EITR(0), newitr);
	}

	return;
}

static void
igc_free_pci_resources(if_ctx_t ctx)
{
#ifdef UNDER_RTSS
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	device_t dev = ctx;
	if (adapter->memory != NULL) {
		bus_release_resource(dev, SYS_RES_MEMORY,
			PCIR_BAR(0), adapter->memory);
		adapter->memory = NULL;
	}
#else
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct igc_rx_queue* que = adapter->rx_queues;
	device_t dev = iflib_get_dev(ctx);

	/* Release all MSI-X queue resources */
	if (adapter->intr_type == IFLIB_INTR_MSIX)
		iflib_irq_free(ctx, &adapter->irq);

	for (int i = 0; i < adapter->rx_num_queues; i++, que++) {
		iflib_irq_free(ctx, &que->que_irq);
	}

	if (adapter->memory != NULL) {
		bus_release_resource(dev, SYS_RES_MEMORY,
			rman_get_rid(adapter->memory), adapter->memory);
		adapter->memory = NULL;
	}

	if (adapter->flash != NULL) {
		bus_release_resource(dev, SYS_RES_MEMORY,
			rman_get_rid(adapter->flash), adapter->flash);
		adapter->flash = NULL;
	}

	if (adapter->ioport != NULL) {
		bus_release_resource(dev, SYS_RES_IOPORT,
			rman_get_rid(adapter->ioport), adapter->ioport);
		adapter->ioport = NULL;
	}
#endif
}

/* Set up MSI or MSI-X */
static int
igc_setup_msix(if_ctx_t ctx)
{
	return (0);
}

/*********************************************************************
 *
 *  Initialize the DMA Coalescing feature
 *
 **********************************************************************/
static void
igc_init_dmac(struct igc_adapter* adapter, u32 pba)
{
	device_t	dev = adapter->dev;
	struct igc_hw* hw = &adapter->hw;
	u32 		dmac, reg = ~IGC_DMACR_DMAC_EN;
	u16		hwm;
	u16		max_frame_size;
	int		status;


	if (adapter->dmac == 0) { /* Disabling it */
		IGC_WRITE_REG(hw, IGC_DMACR, reg);
		return;
	}

	max_frame_size = adapter->hw.mac.max_frame_size =
		ETHERMTU + ETHER_HDR_LEN + ETHERNET_FCS_SIZE;

	/* Set starting threshold */
	IGC_WRITE_REG(hw, IGC_DMCTXTH, 0);

	hwm = 64 * pba - max_frame_size / 16;
	if (hwm < 64 * (pba - 6))
		hwm = 64 * (pba - 6);
	reg = IGC_READ_REG(hw, IGC_FCRTC);
	reg &= ~IGC_FCRTC_RTH_COAL_MASK;
	reg |= ((hwm << IGC_FCRTC_RTH_COAL_SHIFT)
		& IGC_FCRTC_RTH_COAL_MASK);
	IGC_WRITE_REG(hw, IGC_FCRTC, reg);

	dmac = pba - max_frame_size / 512;
	if (dmac < pba - 10)
		dmac = pba - 10;
	reg = IGC_READ_REG(hw, IGC_DMACR);
	reg &= ~IGC_DMACR_DMACTHR_MASK;
	reg |= ((dmac << IGC_DMACR_DMACTHR_SHIFT)
		& IGC_DMACR_DMACTHR_MASK);

	/* transition to L0x or L1 if available..*/
	reg |= (IGC_DMACR_DMAC_EN | IGC_DMACR_DMAC_LX_MASK);

	/* Check if status is 2.5Gb backplane connection
	 * before configuration of watchdog timer, which is
	 * in msec values in 12.8usec intervals
	 * watchdog timer= msec values in 32usec intervals
	 * for non 2.5Gb connection
	 */
	status = IGC_READ_REG(hw, IGC_STATUS);
	if ((status & IGC_STATUS_2P5_SKU) &&
		(!(status & IGC_STATUS_2P5_SKU_OVER)))
		reg |= ((adapter->dmac * 5) >> 6);
	else
		reg |= (adapter->dmac >> 5);

	IGC_WRITE_REG(hw, IGC_DMACR, reg);

	IGC_WRITE_REG(hw, IGC_DMCRTRH, 0);

	/* Set the interval before transition */
	reg = IGC_READ_REG(hw, IGC_DMCTLX);
	reg |= IGC_DMCTLX_DCFLUSH_DIS;

	/*
	** in 2.5Gb connection, TTLX unit is 0.4 usec
	** which is 0x4*2 = 0xA. But delay is still 4 usec
	*/
	status = IGC_READ_REG(hw, IGC_STATUS);
	if ((status & IGC_STATUS_2P5_SKU) &&
		(!(status & IGC_STATUS_2P5_SKU_OVER)))
		reg |= 0xA;
	else
		reg |= 0x4;

	IGC_WRITE_REG(hw, IGC_DMCTLX, reg);

	/* free space in tx packet buffer to wake from DMA coal */
	IGC_WRITE_REG(hw, IGC_DMCTXTH, (IGC_TXPBSIZE -
		(2 * max_frame_size)) >> 6);

	/* make low power state decision controlled by DMA coal */
	reg = IGC_READ_REG(hw, IGC_PCIEMISC);
	reg &= ~IGC_PCIEMISC_LX_DECISION;
	IGC_WRITE_REG(hw, IGC_PCIEMISC, reg);

}

/*********************************************************************
 *
 *  Initialize the hardware to a configuration as specified by the
 *  adapter structure.
 *
 **********************************************************************/
static void
igc_reset(if_ctx_t ctx)
{
	device_t dev = ctx;
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct igc_hw* hw = &adapter->hw;
	u16 rx_buffer_size;
	u32 pba;

	INIT_DEBUGOUT("igc_reset: begin");
	/* Let the firmware know the OS is in control */
	igc_get_hw_control(adapter);

	/*
	 * Packet Buffer Allocation (PBA)
	 * Writing PBA sets the receive portion of the buffer
	 * the remainder is used for the transmit buffer.
	 */
	pba = IGC_PBA_34K;

	INIT_DEBUGOUT1("igc_reset: pba=%dK", pba);

	/*
	 * These parameters control the automatic generation (Tx) and
	 * response (Rx) to Ethernet PAUSE frames.
	 * - High water mark should allow for at least two frames to be
	 *   received after sending an XOFF.
	 * - Low water mark works best when it is very near the high water mark.
	 *   This allows the receiver to restart by sending XON when it has
	 *   drained a bit. Here we use an arbitrary value of 1500 which will
	 *   restart after one full frame is pulled from the buffer. There
	 *   could be several smaller frames in the buffer and if so they will
	 *   not trigger the XON until their total number reduces the buffer
	 *   by 1500.
	 * - The pause time is fairly large at 1000 x 512ns = 512 usec.
	 */
	rx_buffer_size = (pba & 0xffff) << 10;
	hw->fc.high_water = rx_buffer_size -
		roundup2(adapter->hw.mac.max_frame_size, 1024);
	/* 16-byte granularity */
	hw->fc.low_water = hw->fc.high_water - 16;

	if (adapter->fc) /* locally set flow control value? */
		hw->fc.requested_mode = adapter->fc;
	else
		hw->fc.requested_mode = igc_fc_full;

	hw->fc.pause_time = IGC_FC_PAUSE_TIME;

	hw->fc.send_xon = true;

	/* Issue a global reset */
	igc_reset_hw(hw);
	IGC_WRITE_REG(hw, IGC_WUC, 0);

	/* and a re-init */
	if (igc_init_hw(hw) < 0) {
		device_printf(dev, "Hardware Initialization Failed\n");
		return;
	}

	/* Setup DMA Coalescing */
	igc_init_dmac(adapter, pba);

	IGC_WRITE_REG(hw, IGC_VET, ETHERTYPE_VLAN);
	igc_get_phy_info(hw);
	igc_check_for_link(hw);
	INIT_DEBUGOUT("igc_reset: end");
}

/*
 * Initialise the RSS mapping for NICs that support multiple transmit/
 * receive rings.
 */

#define RSSKEYLEN 10
static void
igc_initialize_rss_mapping(struct igc_adapter* adapter)
{
	struct igc_hw* hw = &adapter->hw;
	int i;
	int queue_id = 0;
	u32 reta;
	u32 rss_key[RSSKEYLEN], mrqc, shift = 0;

	/*
	 * The redirection table controls which destination
	 * queue each bucket redirects traffic to.
	 * Each DWORD represents four queues, with the LSB
	 * being the first queue in the DWORD.
	 *
	 * This just allocates buckets to queues using round-robin
	 * allocation.
	 *
	 * NOTE: It Just Happens to line up with the default
	 * RSS allocation method.
	 */

	 /* Warning FM follows */
	reta = 0;
	for (i = 0; i < 128; i++) {
#ifdef RSS
		queue_id = rss_get_indirection_to_bucket(i);
		/*
		 * If we have more queues than buckets, we'll
		 * end up mapping buckets to a subset of the
		 * queues.
		 *
		 * If we have more buckets than queues, we'll
		 * end up instead assigning multiple buckets
		 * to queues.
		 *
		 * Both are suboptimal, but we need to handle
		 * the case so we don't go out of bounds
		 * indexing arrays and such.
		 */
		queue_id = queue_id % adapter->rx_num_queues;
#else

		queue_id = (i % adapter->rx_num_queues);
#endif
		/* Adjust if required */
		queue_id = queue_id << shift;

		/*
		 * The low 8 bits are for hash value (n+0);
		 * The next 8 bits are for hash value (n+1), etc.
		 */
		reta = reta >> 8;
		reta = reta | (((uint32_t)queue_id) << 24);
		if ((i & 3) == 3) {
			IGC_WRITE_REG(hw, IGC_RETA(i >> 2), reta);
			reta = 0;
		}
	}

	/* Now fill in hash table */

	/*
	 * MRQC: Multiple Receive Queues Command
	 * Set queuing to RSS control, number depends on the device.
	 */
	mrqc = IGC_MRQC_ENABLE_RSS_4Q;

#ifdef RSS
	/* XXX ew typecasting */
	rss_getkey((uint8_t*)&rss_key);
#else
	//arc4rand(&rss_key, sizeof(rss_key), 0);
	rss_key[i] = rand();
	RtPrintf(" igc_initialize_rss_mapping Rand value 0x%x \n ", rss_key[i]);
#endif
	for (i = 0; i < RSSKEYLEN; i++)
		IGC_WRITE_REG_ARRAY(hw, IGC_RSSRK(0), i, rss_key[i]);

	/*
	 * Configure the RSS fields to hash upon.
	 */
	mrqc |= (IGC_MRQC_RSS_FIELD_IPV4 |
		IGC_MRQC_RSS_FIELD_IPV4_TCP);
	mrqc |= (IGC_MRQC_RSS_FIELD_IPV6 |
		IGC_MRQC_RSS_FIELD_IPV6_TCP);
	mrqc |= (IGC_MRQC_RSS_FIELD_IPV4_UDP |
		IGC_MRQC_RSS_FIELD_IPV6_UDP);
	mrqc |= (IGC_MRQC_RSS_FIELD_IPV6_UDP_EX |
		IGC_MRQC_RSS_FIELD_IPV6_TCP_EX);

	IGC_WRITE_REG(hw, IGC_MRQC, mrqc);
}
#ifndef UNDER_RTSS
/*********************************************************************
 *
 *  Setup networking device structure and register interface media.
 *
 **********************************************************************/
static int
igc_setup_interface(if_ctx_t ctx)
{
	struct ifnet* ifp = iflib_get_ifp(ctx);
	struct igc_adapter* adapter = iflib_get_softc(ctx);
#if 0 //ToDo
	if_softc_ctx_t scctx = adapter->shared;

	INIT_DEBUGOUT("igc_setup_interface: begin");

	/* Single Queue */
	if (adapter->tx_num_queues == 1) {
		if_setsendqlen(ifp, scctx->isc_ntxd[0] - 1);
		if_setsendqready(ifp);
	}

	/*
	 * Specify the media types supported by this adapter and register
	 * callbacks to update media and link information
	 */
	ifmedia_add(adapter->media, IFM_ETHER | IFM_10_T, 0, NULL);
	ifmedia_add(adapter->media, IFM_ETHER | IFM_10_T | IFM_FDX, 0, NULL);
	ifmedia_add(adapter->media, IFM_ETHER | IFM_100_TX, 0, NULL);
	ifmedia_add(adapter->media, IFM_ETHER | IFM_100_TX | IFM_FDX, 0, NULL);
	ifmedia_add(adapter->media, IFM_ETHER | IFM_1000_T | IFM_FDX, 0, NULL);
	ifmedia_add(adapter->media, IFM_ETHER | IFM_1000_T, 0, NULL);
	ifmedia_add(adapter->media, IFM_ETHER | IFM_2500_T, 0, NULL);

	ifmedia_add(adapter->media, IFM_ETHER | IFM_AUTO, 0, NULL);
	ifmedia_set(adapter->media, IFM_ETHER | IFM_AUTO);
#endif
	return (0);
}

static int
igc_if_tx_queues_alloc(if_ctx_t ctx, caddr_t* vaddrs, uint64_t* paddrs, int ntxqs, int ntxqsets)
{
#if 0 //ToDo
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	if_softc_ctx_t scctx = adapter->shared;
	int error = IGC_SUCCESS;
	struct igc_tx_queue* que;
	int i, j;

	MPASS(adapter->tx_num_queues > 0);
	MPASS(adapter->tx_num_queues == ntxqsets);

	/* First allocate the top level queue structs */
	if (!(adapter->tx_queues =
		(struct igc_tx_queue*)malloc(sizeof(struct igc_tx_queue) *
			adapter->tx_num_queues, M_DEVBUF, M_NOWAIT | M_ZERO))) {
		device_printf(iflib_get_dev(ctx), "Unable to allocate queue memory\n");
		return(ENOMEM);
	}

	for (i = 0, que = adapter->tx_queues; i < adapter->tx_num_queues; i++, que++) {
		/* Set up some basics */

		struct tx_ring* txr = &que->txr;
		txr->adapter = que->adapter = adapter;
		que->me = txr->me = i;

		/* Allocate report status array */
		if (!(txr->tx_rsq = (qidx_t*)malloc(sizeof(qidx_t) * scctx->isc_ntxd[0], M_DEVBUF, M_NOWAIT | M_ZERO))) {
			device_printf(iflib_get_dev(ctx), "failed to allocate rs_idxs memory\n");
			error = ENOMEM;
			goto fail;
		}
		for (j = 0; j < scctx->isc_ntxd[0]; j++)
			txr->tx_rsq[j] = QIDX_INVALID;
		/* get the virtual and physical address of the hardware queues */
		txr->tx_base = (struct igc_tx_desc*)vaddrs[i * ntxqs];
		txr->tx_paddr = paddrs[i * ntxqs];
	}

	if (bootverbose)
		device_printf(iflib_get_dev(ctx),
			"allocated for %d tx_queues\n", adapter->tx_num_queues);
	return (0);
fail:
	igc_if_queues_free(ctx);
	return (error);
#endif
	return 0;
}

static int
igc_if_rx_queues_alloc(if_ctx_t ctx, caddr_t* vaddrs, uint64_t* paddrs, int nrxqs, int nrxqsets)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	int error = IGC_SUCCESS;
	struct igc_rx_queue* que;
	int i;

#if 0 //ToDo
	MPASS(adapter->rx_num_queues > 0);
	MPASS(adapter->rx_num_queues == nrxqsets);

	/* First allocate the top level queue structs */
	if (!(adapter->rx_queues =
		(struct igc_rx_queue*)malloc(sizeof(struct igc_rx_queue) *
			adapter->rx_num_queues, M_DEVBUF, M_NOWAIT | M_ZERO))) {
		device_printf(iflib_get_dev(ctx), "Unable to allocate queue memory\n");
		error = ENOMEM;
		goto fail;
	}

	for (i = 0, que = adapter->rx_queues; i < nrxqsets; i++, que++) {
		/* Set up some basics */
		struct rx_ring* rxr = &que->rxr;
		rxr->adapter = que->adapter = adapter;
		rxr->que = que;
		que->me = rxr->me = i;

		/* get the virtual and physical address of the hardware queues */
		rxr->rx_base = (union igc_rx_desc_extended*)vaddrs[i * nrxqs];
		rxr->rx_paddr = paddrs[i * nrxqs];
	}

	if (bootverbose)
		device_printf(iflib_get_dev(ctx),
			"allocated for %d rx_queues\n", adapter->rx_num_queues);

	return (0);
fail:
	igc_if_queues_free(ctx);
	return (error);
#endif
	return 0;
}

static void
igc_if_queues_free(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct igc_tx_queue* tx_que = adapter->tx_queues;
	struct igc_rx_queue* rx_que = adapter->rx_queues;

	if (tx_que != NULL) {
		for (int i = 0; i < adapter->tx_num_queues; i++, tx_que++) {
			struct tx_ring* txr = &tx_que->txr;
			if (txr->tx_rsq == NULL)
				break;

			free(txr->tx_rsq, M_DEVBUF);
			txr->tx_rsq = NULL;
		}
		free(adapter->tx_queues, M_DEVBUF);
		adapter->tx_queues = NULL;
	}

	if (rx_que != NULL) {
		free(adapter->rx_queues, M_DEVBUF);
		adapter->rx_queues = NULL;
	}

	igc_release_hw_control(adapter);

	if (adapter->mta != NULL) {
		free(adapter->mta, M_DEVBUF);
	}

}
#endif
/*********************************************************************
 *
 *  Enable transmit unit.
 *
 **********************************************************************/
static void
igc_initialize_transmit_unit(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	DRVDEV* pCardInfo = (DRVDEV*)ctx;

	struct igc_hw* hw = &adapter->hw;
	u32 tctl, txdctl = 0;
	unsigned __int64 physicalAddr;

	INIT_DEBUGOUT("igc_initialize_transmit_unit: begin");

	// disable the transmit path
	//IGC_WRITE_REG(&adapter->hw, IGC_TCTL, 0);

	//for (int i = 0; i < adapter->tx_num_queues; i++, txr++) {
	for (int QueNum = 0; QueNum < pCardInfo->numberOfTxNicQueues; QueNum++)
	{
		//IGC_WRITE_REG(hw, IGC_TXDCTL(QueNum), 0); //RTX port
		//u64 bus_addr;
		//caddr_t offp, endp;

		////que = &adapter->tx_queues[i];
		//que = &pCardInfo->adapterQ[i].TxDescriptors;
		//txr = &que->txr;
		//bus_addr = txr->tx_paddr;

		///* Clear checksum offload context. */
		//offp = (caddr_t)&txr->csum_flags;
		//endp = (caddr_t)(txr + 1);
		//bzero(offp, endp - offp);

		physicalAddr = (unsigned __int64)pCardInfo->adapterQ[QueNum].TxDescDMA.QuadPart;

		/* Base and Len of TX Ring */
		IGC_WRITE_REG(hw, IGC_TDLEN(QueNum),
			pCardInfo->adapterQ[QueNum].NumTxDescriptors * sizeof(IGC_ADV_TRANSMIT_DESCRIPTOR));
		IGC_WRITE_REG(hw, IGC_TDBAH(QueNum),
			(physicalAddr >> 32) & 0xFFFFFFFF);
		IGC_WRITE_REG(hw, IGC_TDBAL(QueNum), 
			physicalAddr & 0xFFFFFFFF);
		
		/* Init the HEAD/TAIL indices */
		IGC_WRITE_REG(hw, IGC_TDT(QueNum), 0);
		IGC_WRITE_REG(hw, IGC_TDH(QueNum), 0);
		
		//HW_DEBUGOUT2("Base = %x, Length = %x\n",
		//	IGC_READ_REG(&adapter->hw, IGC_TDBAL(QueNum)),
		//	IGC_READ_REG(&adapter->hw, IGC_TDLEN(QueNum)));

		// Program the Transmit Control Register first
		txdctl = IGC_READ_REG(hw, IGC_TCTL);

		txdctl = 0; /* clear txdctl */
		txdctl |= 0x1f; /* PTHRESH */
		txdctl |= 1 << 8; /* HTHRESH */
		txdctl |= 1 << 16;/* WTHRESH */
		txdctl |= 1 << 22; /* Reserved bit 22 must always be 1 */
		txdctl |= IGC_TXDCTL_GRAN;
		txdctl |= 1 << 25; /* LWTHRESH */
		IGC_WRITE_REG(hw, IGC_TXDCTL(QueNum), txdctl);
	}

	/* Program the Transmit Control Register */
	tctl = IGC_READ_REG(&adapter->hw, IGC_TCTL);
	tctl &= ~IGC_TCTL_CT;
	tctl |= (IGC_TCTL_PSP | IGC_TCTL_RTLC | IGC_TCTL_EN |
		(IGC_COLLISION_THRESHOLD << IGC_CT_SHIFT));

	/* This write will effectively turn on the transmit unit. */
	IGC_WRITE_REG(&adapter->hw, IGC_TCTL, tctl);
	INIT_DEBUGOUT("igc_initialize_transmit_unit: end");
}

/*********************************************************************
 *
 *  Enable receive unit.
 *
 **********************************************************************/

static void
igc_initialize_receive_unit(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	DRVDEV* pCardInfo = (DRVDEV*)ctx;
	struct igc_hw* hw = &adapter->hw;

	u32 psize, rctl, rxcsum, srrctl = 0;
	unsigned __int64 physicalAddr;

	INIT_DEBUGOUT("igc_initialize_receive_units: begin");

	/*
	 * Make sure receives are disabled while setting
	 * up the descriptor ring
	 */
	rctl = IGC_READ_REG(hw, IGC_RCTL);
	IGC_WRITE_REG(hw, IGC_RCTL, rctl & ~IGC_RCTL_EN);

	/* Setup the Receive Control Register */
	rctl &= ~(3 << IGC_RCTL_MO_SHIFT);
#ifndef UNDER_RTSS
	rctl |= IGC_RCTL_EN | IGC_RCTL_BAM |
		IGC_RCTL_LBM_NO | IGC_RCTL_RDMTS_HALF |
		(hw->mac.mc_filter_type << IGC_RCTL_MO_SHIFT);
#else
	rctl |= IGC_RCTL_EN | IGC_RCTL_BAM |
		IGC_RCTL_LBM_NO | IGC_RCTL_RDMTS_HALF ;
#endif
	/* Do not store bad packets */
	rctl &= ~IGC_RCTL_SBP;

	/* Enable Long Packet receive */
	if (pCardInfo->Mtu > ETHERMTU)
		rctl |= IGC_RCTL_LPE;
	else
		rctl &= ~IGC_RCTL_LPE;

	/* Strip the CRC */
	if (!igc_disable_crc_stripping)
		rctl |= IGC_RCTL_SECRC;

	/*
	 * Set the interrupt throttling rate. Value is calculated
	 * as DEFAULT_ITR = 1/(MAX_INTS_PER_SEC * 256ns)
	 */
	IGC_WRITE_REG(hw, IGC_ITR, DEFAULT_ITR);

	rxcsum = IGC_READ_REG(hw, IGC_RXCSUM);
#ifndef UNDER_RTSS
	if (if_getcapenable(ifp) & IFCAP_RXCSUM) {
#else
	if (pCardInfo->RxChecksum == 1) {
		rxcsum |= IGC_RXCSUM_CRCOFL;
		if (pCardInfo->numberOfTxNicQueues > 1)
			rxcsum |= IGC_RXCSUM_PCSD;
		else
			rxcsum |= IGC_RXCSUM_IPPCSE;
	}
	else {
		if (pCardInfo->numberOfTxNicQueues > 1)
			rxcsum |= IGC_RXCSUM_PCSD;
		else
			rxcsum &= ~IGC_RXCSUM_TUOFL;
	}
	IGC_WRITE_REG(hw, IGC_RXCSUM, rxcsum);
#endif
#ifndef UNDER_RTSS
	if (pCardInfo->numberOfRxNicQueues > 1)
		igc_initialize_rss_mapping(adapter);
#endif

	if (pCardInfo->Mtu > ETHERMTU) {
		/* Set maximum packet len. pCardInfo->RxBufferSize already multiple of 1K */
		srrctl |= pCardInfo->RxBufferSize >> IGC_SRRCTL_BSIZEPKT_SHIFT;
		psize = adapter->hw.mac.max_frame_size;
#ifndef UNDER_RTSS
		/* are we on a vlan? */
		if (ifp->if_vlantrunk != NULL)
			psize += VLAN_TAG_SIZE;
#endif
		IGC_WRITE_REG(&adapter->hw, IGC_RLPML, psize);
	}
	else {
		srrctl |= 2048 >> IGC_SRRCTL_BSIZEPKT_SHIFT;
		rctl |= IGC_RCTL_SZ_2048;
	}

	/*
	 * If TX flow control is disabled and there's >1 queue defined,
	 * enable DROP.
	 *
	 * This drops frames rather than hanging the RX MAC for all queues.
	 */
	if ((pCardInfo->numberOfRxNicQueues > 1) &&
		(adapter->fc == igc_fc_none))
	{
		srrctl |= IGC_SRRCTL_DROP_EN;
	}

	/* Setup the Base and Length of the Rx Descriptor Rings */
	for (int QueNum = 0; QueNum < pCardInfo->numberOfRxNicQueues; QueNum++) {
		u32 rxdctl;
		u32 rdt = pCardInfo->adapterQ[QueNum].NumRxDescriptors - 1; /* default */

#ifdef notyet
		/* Configure for header split? -- ignore for now */
		rxr->hdr_split = igc_header_split;
#else
		srrctl |= IGC_SRRCTL_DESCTYPE_ADV_ONEBUF;
#endif

		physicalAddr = (unsigned __int64)pCardInfo->adapterQ[QueNum].RxDescDMA.QuadPart;
		
		IGC_WRITE_REG(hw, IGC_RDLEN(QueNum), 
			pCardInfo->adapterQ[QueNum].NumRxDescriptors * sizeof(IGC_ADV_RECEIVE_DESCRIPTOR));
		
		IGC_WRITE_REG(hw, IGC_RDBAH(QueNum), (u32)(physicalAddr >> 32));
		IGC_WRITE_REG(hw, IGC_RDBAL(QueNum), (u32)physicalAddr);

		IGC_WRITE_REG(hw, IGC_SRRCTL(QueNum), srrctl);
		/* Setup the Head and Tail Descriptor Pointers */
		IGC_WRITE_REG(hw, IGC_RDH(QueNum), 0);
		IGC_WRITE_REG(hw, IGC_RDT(QueNum), 0);
	
		/* Enable this Queue */
		rxdctl = IGC_READ_REG(hw, IGC_RXDCTL(QueNum));
		rxdctl |= IGC_RXDCTL_QUEUE_ENABLE;
		rxdctl &= 0xFFF00000;
		rxdctl |= IGC_RX_PTHRESH;
		rxdctl |= IGC_RX_HTHRESH << 8;
		rxdctl |= IGC_RX_WTHRESH << 16;
		IGC_WRITE_REG(hw, IGC_RXDCTL(QueNum), rxdctl);

		//RtPrintf("Receive Base high = 0x%x Base low = %x, Length = %x\n", 
		//	  IGC_READ_REG(&adapter->hw, IGC_RDBAH(QueNum)),
		//	  IGC_READ_REG(&adapter->hw, IGC_RDBAL(QueNum)),
		//	  IGC_READ_REG(&adapter->hw, IGC_RDLEN(QueNum)));
		
	   //RTX Port update Tail with number of descriptors available
	   IGC_WRITE_REG(hw, IGC_RDT(QueNum), rdt);
	}

	/* Make sure VLAN Filters are off */
	rctl &= ~IGC_RCTL_VFE;
	
	/* Write out the settings */
	IGC_WRITE_REG(hw, IGC_RCTL, rctl);

	INIT_DEBUGOUT("igc_initialize_receive_units: end");

	return;
}
#ifndef UNDER_RTSS
static void
igc_if_vlan_register(if_ctx_t ctx, u16 vtag)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	u32 index, bit;

	index = (vtag >> 5) & 0x7F;
	bit = vtag & 0x1F;
	adapter->shadow_vfta[index] |= (1 << bit);
	++adapter->num_vlans;
}

static void
igc_if_vlan_unregister(if_ctx_t ctx, u16 vtag)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	u32 index, bit;

	index = (vtag >> 5) & 0x7F;
	bit = vtag & 0x1F;
	adapter->shadow_vfta[index] &= ~(1 << bit);
	--adapter->num_vlans;
}

static void
igc_setup_vlan_hw_support(struct igc_adapter* adapter)
{
	struct igc_hw* hw = &adapter->hw;
	u32 reg;

	/*
	 * We get here thru init_locked, meaning
	 * a soft reset, this has already cleared
	 * the VFTA and other state, so if there
	 * have been no vlan's registered do nothing.
	 */
	if (adapter->num_vlans == 0)
		return;

	/*
	 * A soft reset zero's out the VFTA, so
	 * we need to repopulate it now.
	 */
	for (int i = 0; i < IGC_VFTA_SIZE; i++)
		if (adapter->shadow_vfta[i] != 0)
			IGC_WRITE_REG_ARRAY(hw, IGC_VFTA,
				i, adapter->shadow_vfta[i]);

	reg = IGC_READ_REG(hw, IGC_CTRL);
	reg |= IGC_CTRL_VME;
	IGC_WRITE_REG(hw, IGC_CTRL, reg);

	/* Enable the Filter Table */
	reg = IGC_READ_REG(hw, IGC_RCTL);
	reg &= ~IGC_RCTL_CFIEN;
	reg |= IGC_RCTL_VFE;
	IGC_WRITE_REG(hw, IGC_RCTL, reg);
}
#endif
static void
igc_if_intr_enable(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct igc_hw* hw = &adapter->hw;
	u32 mask = IMS_ENABLE_MASK;
#if 0
	if (adapter->intr_type == IFLIB_INTR_MSIX) {
		//enabling extended interrupts has performance impact
		//(IGC_EICR_RX_QUEUE | IGC_EICR_TX_QUEUE);
		mask = 0;
		IGC_WRITE_REG(hw, IGC_EIAC, mask);
		IGC_WRITE_REG(hw, IGC_EIAM, mask);
		IGC_WRITE_REG(hw, IGC_EIMS, mask);
		IGC_WRITE_REG(hw, IGC_IMS, IGC_IMS_LSC);
#else
if (adapter->intr_type == IFLIB_INTR_MSIX) {
		mask = (u32) (adapter->que_mask | adapter->link_mask);
		IGC_WRITE_REG(hw, IGC_EIAC, mask);
		IGC_WRITE_REG(hw, IGC_EIAM, mask);
		IGC_WRITE_REG(hw, IGC_EIMS, mask);
		IGC_WRITE_REG(hw, IGC_IMS, IGC_IMS_LSC);
#endif		
		IGC_WRITE_REG(hw, IGC_EIMS, IMS_ENABLE_MASK);

		IGC_WRITE_FLUSH(hw);
	}
	else
		IGC_WRITE_REG(hw, IGC_IMS, IMS_ENABLE_MASK);
	IGC_WRITE_FLUSH(hw);
}

static void
igc_if_intr_disable(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct igc_hw* hw = &adapter->hw;

	if (adapter->intr_type == IFLIB_INTR_MSIX) {
		IGC_WRITE_REG(hw, IGC_EIMC, 0xffffffff);
		IGC_WRITE_REG(hw, IGC_EIAC, 0);
	}
	IGC_WRITE_REG(hw, IGC_IMC, 0xffffffff);
	IGC_WRITE_FLUSH(hw);
}

/*
 * igc_get_hw_control sets the {CTRL_EXT|FWSM}:DRV_LOAD bit.
 * For ASF and Pass Through versions of f/w this means
 * that the driver is loaded. For AMT version type f/w
 * this means that the network i/f is open.
 */

static void
igc_get_hw_control(struct igc_adapter* adapter)
{
	u32 ctrl_ext;

	if (adapter->vf_ifp)
		return;

	ctrl_ext = IGC_READ_REG(&adapter->hw, IGC_CTRL_EXT);
	IGC_WRITE_REG(&adapter->hw, IGC_CTRL_EXT,
		ctrl_ext | IGC_CTRL_EXT_DRV_LOAD);
}

/*
 * igc_release_hw_control resets {CTRL_EXT|FWSM}:DRV_LOAD bit.
 * For ASF and Pass Through versions of f/w this means that
 * the driver is no longer loaded. For AMT versions of the
 * f/w this means that the network i/f is closed.
 */
static void
igc_release_hw_control(struct igc_adapter* adapter)
{
	u32 ctrl_ext;

	ctrl_ext = IGC_READ_REG(&adapter->hw, IGC_CTRL_EXT);
	IGC_WRITE_REG(&adapter->hw, IGC_CTRL_EXT,
		ctrl_ext & ~IGC_CTRL_EXT_DRV_LOAD);
	return;
}

static int
igc_is_valid_ether_addr(u8* addr)
{
	char zero_addr[6] = { 0, 0, 0, 0, 0, 0 };

	if ((addr[0] & 1) || (!memcmp(addr, zero_addr, ETHER_ADDR_LEN))) {
		return (false);
	}

	return (true);
}

/*
** Parse the interface capabilities with regard
** to both system management and wake-on-lan for
** later use.
*/
static void
igc_get_wakeup(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	u32 eeprom_data = 0, apme_mask;

	apme_mask = IGC_WUC_APME;
	eeprom_data = IGC_READ_REG(&adapter->hw, IGC_WUC);

	if (eeprom_data & apme_mask)
		adapter->wol = IGC_WUFC_LNKC;
}


/*
 * Enable PCI Wake On Lan capability
 */
static void
igc_enable_wakeup(if_ctx_t ctx)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	device_t dev = ctx;
#if 0 //ToDo
	if_t ifp = iflib_get_ifp(ctx);
	int error = 0;
	u32 pmc, ctrl, rctl;
	u16 status;

	if (pci_find_cap(dev, PCIY_PMG, &pmc) != 0)
		return;

	/*
	 * Determine type of Wakeup: note that wol
	 * is set with all bits on by default.
	 */
	if ((if_getcapenable(ifp) & IFCAP_WOL_MAGIC) == 0)
		adapter->wol &= ~IGC_WUFC_MAG;

	if ((if_getcapenable(ifp) & IFCAP_WOL_UCAST) == 0)
		adapter->wol &= ~IGC_WUFC_EX;

	if ((if_getcapenable(ifp) & IFCAP_WOL_MCAST) == 0)
		adapter->wol &= ~IGC_WUFC_MC;
	else {
		rctl = IGC_READ_REG(&adapter->hw, IGC_RCTL);
		rctl |= IGC_RCTL_MPE;
		IGC_WRITE_REG(&adapter->hw, IGC_RCTL, rctl);
	}

	if (!(adapter->wol & (IGC_WUFC_EX | IGC_WUFC_MAG | IGC_WUFC_MC)))
		goto pme;

	/* Advertise the wakeup capability */
	ctrl = IGC_READ_REG(&adapter->hw, IGC_CTRL);
	ctrl |= IGC_CTRL_ADVD3WUC;
	IGC_WRITE_REG(&adapter->hw, IGC_CTRL, ctrl);

	/* Enable wakeup by the MAC */
	IGC_WRITE_REG(&adapter->hw, IGC_WUC, IGC_WUC_PME_EN);
	IGC_WRITE_REG(&adapter->hw, IGC_WUFC, adapter->wol);

pme:
	status = pci_read_config(dev, pmc + PCIR_POWER_STATUS, 2);
	status &= ~(PCIM_PSTAT_PME | PCIM_PSTAT_PMEENABLE);
	if (!error && (if_getcapenable(ifp) & IFCAP_WOL))
		status |= PCIM_PSTAT_PME | PCIM_PSTAT_PMEENABLE;
	pci_write_config(dev, pmc + PCIR_POWER_STATUS, status, 2);
#endif
	return;
}

/**********************************************************************
 *
 *  Update the board statistics counters.
 *
 **********************************************************************/
static void
igc_update_stats_counters(struct igc_adapter* adapter)
{
#ifndef IGC_PRINTSTATS
	u64 prev_xoffrxc = adapter->stats.xoffrxc;

	adapter->stats.crcerrs += IGC_READ_REG(&adapter->hw, IGC_CRCERRS);
	adapter->stats.mpc += IGC_READ_REG(&adapter->hw, IGC_MPC);
	adapter->stats.scc += IGC_READ_REG(&adapter->hw, IGC_SCC);
	adapter->stats.ecol += IGC_READ_REG(&adapter->hw, IGC_ECOL);

	adapter->stats.mcc += IGC_READ_REG(&adapter->hw, IGC_MCC);
	adapter->stats.latecol += IGC_READ_REG(&adapter->hw, IGC_LATECOL);
	adapter->stats.colc += IGC_READ_REG(&adapter->hw, IGC_COLC);
	adapter->stats.colc += IGC_READ_REG(&adapter->hw, IGC_RERC);
	adapter->stats.dc += IGC_READ_REG(&adapter->hw, IGC_DC);
	adapter->stats.rlec += IGC_READ_REG(&adapter->hw, IGC_RLEC);
	adapter->stats.xonrxc += IGC_READ_REG(&adapter->hw, IGC_XONRXC);
	adapter->stats.xontxc += IGC_READ_REG(&adapter->hw, IGC_XONTXC);
	adapter->stats.xoffrxc += IGC_READ_REG(&adapter->hw, IGC_XOFFRXC);
	/*
	 * For watchdog management we need to know if we have been
	 * paused during the last interval, so capture that here.
	 */
	//if (adapter->stats.xoffrxc != prev_xoffrxc)
	//	adapter->shared->isc_pause_frames = 1;
	adapter->stats.xofftxc += IGC_READ_REG(&adapter->hw, IGC_XOFFTXC);
	adapter->stats.fcruc += IGC_READ_REG(&adapter->hw, IGC_FCRUC);
	adapter->stats.prc64 += IGC_READ_REG(&adapter->hw, IGC_PRC64);
	adapter->stats.prc127 += IGC_READ_REG(&adapter->hw, IGC_PRC127);
	adapter->stats.prc255 += IGC_READ_REG(&adapter->hw, IGC_PRC255);
	adapter->stats.prc511 += IGC_READ_REG(&adapter->hw, IGC_PRC511);
	adapter->stats.prc1023 += IGC_READ_REG(&adapter->hw, IGC_PRC1023);
	adapter->stats.prc1522 += IGC_READ_REG(&adapter->hw, IGC_PRC1522);
	adapter->stats.tlpic += IGC_READ_REG(&adapter->hw, IGC_TLPIC);
	adapter->stats.rlpic += IGC_READ_REG(&adapter->hw, IGC_RLPIC);
	adapter->stats.gprc += IGC_READ_REG(&adapter->hw, IGC_GPRC);
	adapter->stats.bprc += IGC_READ_REG(&adapter->hw, IGC_BPRC);
	adapter->stats.mprc += IGC_READ_REG(&adapter->hw, IGC_MPRC);
	adapter->stats.gptc += IGC_READ_REG(&adapter->hw, IGC_GPTC);

	/* For the 64-bit byte counters the low dword must be read first. */
	/* Both registers clear on the read of the high dword */

	adapter->stats.gorc += IGC_READ_REG(&adapter->hw, IGC_GORCL) +
		((u64)IGC_READ_REG(&adapter->hw, IGC_GORCH) << 32);
	adapter->stats.gotc += IGC_READ_REG(&adapter->hw, IGC_GOTCL) +
		((u64)IGC_READ_REG(&adapter->hw, IGC_GOTCH) << 32);

	adapter->stats.rnbc += IGC_READ_REG(&adapter->hw, IGC_RNBC);
	adapter->stats.ruc += IGC_READ_REG(&adapter->hw, IGC_RUC);
	adapter->stats.rfc += IGC_READ_REG(&adapter->hw, IGC_RFC);
	adapter->stats.roc += IGC_READ_REG(&adapter->hw, IGC_ROC);
	adapter->stats.rjc += IGC_READ_REG(&adapter->hw, IGC_RJC);

	adapter->stats.tor += IGC_READ_REG(&adapter->hw, IGC_TORH);
	adapter->stats.tot += IGC_READ_REG(&adapter->hw, IGC_TOTH);

	adapter->stats.tpr += IGC_READ_REG(&adapter->hw, IGC_TPR);
	adapter->stats.tpt += IGC_READ_REG(&adapter->hw, IGC_TPT);
	adapter->stats.ptc64 += IGC_READ_REG(&adapter->hw, IGC_PTC64);
	adapter->stats.ptc127 += IGC_READ_REG(&adapter->hw, IGC_PTC127);
	adapter->stats.ptc255 += IGC_READ_REG(&adapter->hw, IGC_PTC255);
	adapter->stats.ptc511 += IGC_READ_REG(&adapter->hw, IGC_PTC511);
	adapter->stats.ptc1023 += IGC_READ_REG(&adapter->hw, IGC_PTC1023);
	adapter->stats.ptc1522 += IGC_READ_REG(&adapter->hw, IGC_PTC1522);
	adapter->stats.mptc += IGC_READ_REG(&adapter->hw, IGC_MPTC);
	adapter->stats.bptc += IGC_READ_REG(&adapter->hw, IGC_BPTC);

	/* Interrupt Counts */
	adapter->stats.iac += IGC_READ_REG(&adapter->hw, IGC_IAC);
	adapter->stats.rxdmtc += IGC_READ_REG(&adapter->hw, IGC_RXDMTC);

	adapter->stats.algnerrc += IGC_READ_REG(&adapter->hw, IGC_ALGNERRC);
	adapter->stats.tncrs += IGC_READ_REG(&adapter->hw, IGC_TNCRS);
	adapter->stats.htdpmc += IGC_READ_REG(&adapter->hw, IGC_HTDPMC);
	adapter->stats.tsctc += IGC_READ_REG(&adapter->hw, IGC_TSCTC);
#else
	igc_print_stats_counters(adapter);
#endif
}
#ifndef UNDER_RTSS
static uint64_t
igc_if_get_counter(if_ctx_t ctx, ift_counter cnt)
{
	struct igc_adapter* adapter = iflib_get_softc(ctx);
	struct ifnet* ifp = iflib_get_ifp(ctx);

	switch (cnt) {
	case IFCOUNTER_COLLISIONS:
		return (adapter->stats.colc);
	case IFCOUNTER_IERRORS:
		return (adapter->dropped_pkts + adapter->stats.rxerrc +
			adapter->stats.crcerrs + adapter->stats.algnerrc +
			adapter->stats.ruc + adapter->stats.roc +
			adapter->stats.mpc + adapter->stats.htdpmc);
	case IFCOUNTER_OERRORS:
		return (adapter->stats.ecol + adapter->stats.latecol +
			adapter->watchdog_events);
	default:
		return (if_get_counter_default(ifp, cnt));
	}
}
#endif
/* igc_if_needs_restart - Tell iflib when the driver needs to be reinitialized
 * @ctx: iflib context
 * @event: event code to check
 *
 * Defaults to returning true for unknown events.
 *
 * @returns true if iflib needs to reinit the interface
 */
static bool
igc_if_needs_restart(if_ctx_t ctx , enum iflib_restart_event event)
{
#if 0 //ToDo
	switch (event) {
	case IFLIB_RESTART_VLAN_CONFIG:
	default:
		return (true);
	}
#endif
	return (true);
}

/* Export a single 32-bit register via a read-only sysctl. */
static int
igc_sysctl_reg_handler(SYSCTL_HANDLER_ARGS)
{
#if 0 //ToDo
	struct igc_adapter* adapter;
	u_int val;

	adapter = oidp->oid_arg1;
	val = IGC_READ_REG(&adapter->hw, oidp->oid_arg2);
	return (sysctl_handle_int(oidp, &val, 0, req));
#endif
	return 0;
}

/*
 * Add sysctl variables, one per statistic, to the system.
 */
static void
igc_add_hw_stats(struct igc_adapter* adapter)
{
#if 0 //ToDo
	device_t dev = iflib_get_dev(adapter->ctx);
	struct igc_tx_queue* tx_que = adapter->tx_queues;
	struct igc_rx_queue* rx_que = adapter->rx_queues;

	struct sysctl_ctx_list* ctx = device_get_sysctl_ctx(dev);
	struct sysctl_oid* tree = device_get_sysctl_tree(dev);
	struct sysctl_oid_list* child = SYSCTL_CHILDREN(tree);
	struct igc_hw_stats* stats = &adapter->stats;

	struct sysctl_oid* stat_node, * queue_node, * int_node;
	struct sysctl_oid_list* stat_list, * queue_list, * int_list;

#define QUEUE_NAME_LEN 32
	char namebuf[QUEUE_NAME_LEN];

	/* Driver Statistics */
	SYSCTL_ADD_ULONG(ctx, child, OID_AUTO, "dropped",
		CTLFLAG_RD, &adapter->dropped_pkts,
		"Driver dropped packets");
	SYSCTL_ADD_ULONG(ctx, child, OID_AUTO, "link_irq",
		CTLFLAG_RD, &adapter->link_irq,
		"Link MSI-X IRQ Handled");
	SYSCTL_ADD_ULONG(ctx, child, OID_AUTO, "rx_overruns",
		CTLFLAG_RD, &adapter->rx_overruns,
		"RX overruns");
	SYSCTL_ADD_ULONG(ctx, child, OID_AUTO, "watchdog_timeouts",
		CTLFLAG_RD, &adapter->watchdog_events,
		"Watchdog timeouts");
	SYSCTL_ADD_PROC(ctx, child, OID_AUTO, "device_control",
		CTLTYPE_UINT | CTLFLAG_RD | CTLFLAG_NEEDGIANT,
		adapter, IGC_CTRL, igc_sysctl_reg_handler, "IU",
		"Device Control Register");
	SYSCTL_ADD_PROC(ctx, child, OID_AUTO, "rx_control",
		CTLTYPE_UINT | CTLFLAG_RD | CTLFLAG_NEEDGIANT,
		adapter, IGC_RCTL, igc_sysctl_reg_handler, "IU",
		"Receiver Control Register");
	SYSCTL_ADD_UINT(ctx, child, OID_AUTO, "fc_high_water",
		CTLFLAG_RD, &adapter->hw.fc.high_water, 0,
		"Flow Control High Watermark");
	SYSCTL_ADD_UINT(ctx, child, OID_AUTO, "fc_low_water",
		CTLFLAG_RD, &adapter->hw.fc.low_water, 0,
		"Flow Control Low Watermark");

	for (int i = 0; i < adapter->tx_num_queues; i++, tx_que++) {
		struct tx_ring* txr = &tx_que->txr;
		snprintf(namebuf, QUEUE_NAME_LEN, "queue_tx_%d", i);
		queue_node = SYSCTL_ADD_NODE(ctx, child, OID_AUTO, namebuf,
			CTLFLAG_RD | CTLFLAG_MPSAFE, NULL, "TX Queue Name");
		queue_list = SYSCTL_CHILDREN(queue_node);

		SYSCTL_ADD_PROC(ctx, queue_list, OID_AUTO, "txd_head",
			CTLTYPE_UINT | CTLFLAG_RD | CTLFLAG_NEEDGIANT, adapter,
			IGC_TDH(txr->me), igc_sysctl_reg_handler, "IU",
			"Transmit Descriptor Head");
		SYSCTL_ADD_PROC(ctx, queue_list, OID_AUTO, "txd_tail",
			CTLTYPE_UINT | CTLFLAG_RD | CTLFLAG_NEEDGIANT, adapter,
			IGC_TDT(txr->me), igc_sysctl_reg_handler, "IU",
			"Transmit Descriptor Tail");
		SYSCTL_ADD_ULONG(ctx, queue_list, OID_AUTO, "tx_irq",
			CTLFLAG_RD, &txr->tx_irq,
			"Queue MSI-X Transmit Interrupts");
	}

	for (int j = 0; j < adapter->rx_num_queues; j++, rx_que++) {
		struct rx_ring* rxr = &rx_que->rxr;
		snprintf(namebuf, QUEUE_NAME_LEN, "queue_rx_%d", j);
		queue_node = SYSCTL_ADD_NODE(ctx, child, OID_AUTO, namebuf,
			CTLFLAG_RD | CTLFLAG_MPSAFE, NULL, "RX Queue Name");
		queue_list = SYSCTL_CHILDREN(queue_node);

		SYSCTL_ADD_PROC(ctx, queue_list, OID_AUTO, "rxd_head",
			CTLTYPE_UINT | CTLFLAG_RD | CTLFLAG_NEEDGIANT, adapter,
			IGC_RDH(rxr->me), igc_sysctl_reg_handler, "IU",
			"Receive Descriptor Head");
		SYSCTL_ADD_PROC(ctx, queue_list, OID_AUTO, "rxd_tail",
			CTLTYPE_UINT | CTLFLAG_RD | CTLFLAG_NEEDGIANT, adapter,
			IGC_RDT(rxr->me), igc_sysctl_reg_handler, "IU",
			"Receive Descriptor Tail");
		SYSCTL_ADD_ULONG(ctx, queue_list, OID_AUTO, "rx_irq",
			CTLFLAG_RD, &rxr->rx_irq,
			"Queue MSI-X Receive Interrupts");
	}

	/* MAC stats get their own sub node */

	stat_node = SYSCTL_ADD_NODE(ctx, child, OID_AUTO, "mac_stats",
		CTLFLAG_RD | CTLFLAG_MPSAFE, NULL, "Statistics");
	stat_list = SYSCTL_CHILDREN(stat_node);

	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "excess_coll",
		CTLFLAG_RD, &stats->ecol,
		"Excessive collisions");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "single_coll",
		CTLFLAG_RD, &stats->scc,
		"Single collisions");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "multiple_coll",
		CTLFLAG_RD, &stats->mcc,
		"Multiple collisions");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "late_coll",
		CTLFLAG_RD, &stats->latecol,
		"Late collisions");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "collision_count",
		CTLFLAG_RD, &stats->colc,
		"Collision Count");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "symbol_errors",
		CTLFLAG_RD, &adapter->stats.symerrs,
		"Symbol Errors");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "sequence_errors",
		CTLFLAG_RD, &adapter->stats.sec,
		"Sequence Errors");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "defer_count",
		CTLFLAG_RD, &adapter->stats.dc,
		"Defer Count");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "missed_packets",
		CTLFLAG_RD, &adapter->stats.mpc,
		"Missed Packets");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "recv_no_buff",
		CTLFLAG_RD, &adapter->stats.rnbc,
		"Receive No Buffers");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "recv_undersize",
		CTLFLAG_RD, &adapter->stats.ruc,
		"Receive Undersize");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "recv_fragmented",
		CTLFLAG_RD, &adapter->stats.rfc,
		"Fragmented Packets Received ");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "recv_oversize",
		CTLFLAG_RD, &adapter->stats.roc,
		"Oversized Packets Received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "recv_jabber",
		CTLFLAG_RD, &adapter->stats.rjc,
		"Recevied Jabber");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "recv_errs",
		CTLFLAG_RD, &adapter->stats.rxerrc,
		"Receive Errors");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "crc_errs",
		CTLFLAG_RD, &adapter->stats.crcerrs,
		"CRC errors");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "alignment_errs",
		CTLFLAG_RD, &adapter->stats.algnerrc,
		"Alignment Errors");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "xon_recvd",
		CTLFLAG_RD, &adapter->stats.xonrxc,
		"XON Received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "xon_txd",
		CTLFLAG_RD, &adapter->stats.xontxc,
		"XON Transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "xoff_recvd",
		CTLFLAG_RD, &adapter->stats.xoffrxc,
		"XOFF Received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "xoff_txd",
		CTLFLAG_RD, &adapter->stats.xofftxc,
		"XOFF Transmitted");

	/* Packet Reception Stats */
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "total_pkts_recvd",
		CTLFLAG_RD, &adapter->stats.tpr,
		"Total Packets Received ");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "good_pkts_recvd",
		CTLFLAG_RD, &adapter->stats.gprc,
		"Good Packets Received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "bcast_pkts_recvd",
		CTLFLAG_RD, &adapter->stats.bprc,
		"Broadcast Packets Received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "mcast_pkts_recvd",
		CTLFLAG_RD, &adapter->stats.mprc,
		"Multicast Packets Received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "rx_frames_64",
		CTLFLAG_RD, &adapter->stats.prc64,
		"64 byte frames received ");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "rx_frames_65_127",
		CTLFLAG_RD, &adapter->stats.prc127,
		"65-127 byte frames received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "rx_frames_128_255",
		CTLFLAG_RD, &adapter->stats.prc255,
		"128-255 byte frames received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "rx_frames_256_511",
		CTLFLAG_RD, &adapter->stats.prc511,
		"256-511 byte frames received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "rx_frames_512_1023",
		CTLFLAG_RD, &adapter->stats.prc1023,
		"512-1023 byte frames received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "rx_frames_1024_1522",
		CTLFLAG_RD, &adapter->stats.prc1522,
		"1023-1522 byte frames received");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "good_octets_recvd",
		CTLFLAG_RD, &adapter->stats.gorc,
		"Good Octets Received");

	/* Packet Transmission Stats */
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "good_octets_txd",
		CTLFLAG_RD, &adapter->stats.gotc,
		"Good Octets Transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "total_pkts_txd",
		CTLFLAG_RD, &adapter->stats.tpt,
		"Total Packets Transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "good_pkts_txd",
		CTLFLAG_RD, &adapter->stats.gptc,
		"Good Packets Transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "bcast_pkts_txd",
		CTLFLAG_RD, &adapter->stats.bptc,
		"Broadcast Packets Transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "mcast_pkts_txd",
		CTLFLAG_RD, &adapter->stats.mptc,
		"Multicast Packets Transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "tx_frames_64",
		CTLFLAG_RD, &adapter->stats.ptc64,
		"64 byte frames transmitted ");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "tx_frames_65_127",
		CTLFLAG_RD, &adapter->stats.ptc127,
		"65-127 byte frames transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "tx_frames_128_255",
		CTLFLAG_RD, &adapter->stats.ptc255,
		"128-255 byte frames transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "tx_frames_256_511",
		CTLFLAG_RD, &adapter->stats.ptc511,
		"256-511 byte frames transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "tx_frames_512_1023",
		CTLFLAG_RD, &adapter->stats.ptc1023,
		"512-1023 byte frames transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "tx_frames_1024_1522",
		CTLFLAG_RD, &adapter->stats.ptc1522,
		"1024-1522 byte frames transmitted");
	SYSCTL_ADD_UQUAD(ctx, stat_list, OID_AUTO, "tso_txd",
		CTLFLAG_RD, &adapter->stats.tsctc,
		"TSO Contexts Transmitted");

	/* Interrupt Stats */

	int_node = SYSCTL_ADD_NODE(ctx, child, OID_AUTO, "interrupts",
		CTLFLAG_RD | CTLFLAG_MPSAFE, NULL, "Interrupt Statistics");
	int_list = SYSCTL_CHILDREN(int_node);

	SYSCTL_ADD_UQUAD(ctx, int_list, OID_AUTO, "asserts",
		CTLFLAG_RD, &adapter->stats.iac,
		"Interrupt Assertion Count");

	SYSCTL_ADD_UQUAD(ctx, int_list, OID_AUTO, "rx_desc_min_thresh",
		CTLFLAG_RD, &adapter->stats.rxdmtc,
		"Rx Desc Min Thresh Count");
#endif
}

/**********************************************************************
 *
 *  This routine provides a way to dump out the adapter eeprom,
 *  often a useful debug/service tool. This only dumps the first
 *  32 words, stuff that matters is in that extent.
 *
 **********************************************************************/
static int
igc_sysctl_nvm_info(SYSCTL_HANDLER_ARGS)
{
#if 0 //ToDo
	struct igc_adapter* adapter = (struct igc_adapter*)arg1;
	int error;
	int result;

	result = -1;
	error = sysctl_handle_int(oidp, &result, 0, req);

	if (error || !req->newptr)
		return (error);

	/*
	 * This value will cause a hex dump of the
	 * first 32 16-bit words of the EEPROM to
	 * the screen.
	 */
	if (result == 1)
		igc_print_nvm_info(adapter);

	return (error);
#endif
	return 0;
}

static void
igc_print_nvm_info(struct igc_adapter* adapter)
{
	u16 eeprom_data;
	int i, j, row = 0;

	/* Its a bit crude, but it gets the job done */
	printf("\nInterface EEPROM Dump:\n");
	printf("Offset\n0x0000  ");
	for (i = 0, j = 0; i < 32; i++, j++) {
		if (j == 8) { /* Make the offset block */
			j = 0; ++row;
			printf("\n0x00%x0  ", row);
		}
		igc_read_nvm(&adapter->hw, i, 1, &eeprom_data);
		printf("%04x ", eeprom_data);
	}
	printf("\n");
}

static int
igc_sysctl_int_delay(SYSCTL_HANDLER_ARGS)
{
#if 0 //ToDo
	struct igc_int_delay_info* info;
	struct igc_adapter* adapter;
	u32 regval;
	int error, usecs, ticks;

	info = (struct igc_int_delay_info*)arg1;
	usecs = info->value;
	error = sysctl_handle_int(oidp, &usecs, 0, req);
	if (error != 0 || req->newptr == NULL)
		return (error);
	if (usecs < 0 || usecs > IGC_TICKS_TO_USECS(65535))
		return (EINVAL);
	info->value = usecs;
	ticks = IGC_USECS_TO_TICKS(usecs);
	if (info->offset == IGC_ITR)	/* units are 256ns here */
		ticks *= 4;

	adapter = info->adapter;

	regval = IGC_READ_OFFSET(&adapter->hw, info->offset);
	regval = (regval & ~0xffff) | (ticks & 0xffff);
	/* Handle a few special cases. */
	switch (info->offset) {
	case IGC_RDTR:
		break;
	case IGC_TIDV:
		if (ticks == 0) {
			adapter->txd_cmd &= ~IGC_TXD_CMD_IDE;
			/* Don't write 0 into the TIDV register. */
			regval++;
		}
		else
			adapter->txd_cmd |= IGC_TXD_CMD_IDE;
		break;
	}
	IGC_WRITE_OFFSET(&adapter->hw, info->offset, regval);
#endif
	return (0);
}

static void
igc_add_int_delay_sysctl(struct igc_adapter* adapter, const char* name,
	const char* description, struct igc_int_delay_info* info,
	int offset, int value)
{
#if 0 //ToDo
	info->adapter = adapter;
	info->offset = offset;
	info->value = value;
	SYSCTL_ADD_PROC(device_get_sysctl_ctx(adapter->dev),
		SYSCTL_CHILDREN(device_get_sysctl_tree(adapter->dev)),
		OID_AUTO, name, CTLTYPE_INT | CTLFLAG_RW | CTLFLAG_NEEDGIANT,
		info, 0, igc_sysctl_int_delay, "I", description);
#endif
}

/*
 * Set flow control using sysctl:
 * Flow control values:
 *      0 - off
 *      1 - rx pause
 *      2 - tx pause
 *      3 - full
 */
static int
igc_set_flowcntl(SYSCTL_HANDLER_ARGS)
{
#if 0 //ToDo
	int error;
	static int input = 3; /* default is full */

	struct igc_adapter* adapter = (struct igc_adapter*)arg1;

	error = sysctl_handle_int(oidp, &input, 0, req);

	if ((error) || (req->newptr == NULL))
		return (error);

	if (input == adapter->fc) /* no change? */
		return (error);

	switch (input) {
	case igc_fc_rx_pause:
	case igc_fc_tx_pause:
	case igc_fc_full:
	case igc_fc_none:
		adapter->hw.fc.requested_mode = input;
		adapter->fc = input;
		break;
	default:
		/* Do nothing */
		return (error);
	}

	adapter->hw.fc.current_mode = adapter->hw.fc.requested_mode;
	igc_force_mac_fc(&adapter->hw);
	return (error);
#endif
	return 0;
}

/*
 * Manage Energy Efficient Ethernet:
 * Control values:
 *     0/1 - enabled/disabled
 */
static int
igc_sysctl_eee(SYSCTL_HANDLER_ARGS)
{
#if 0 //ToDo
	struct igc_adapter* adapter = (struct igc_adapter*)arg1;
	int error, value;

	value = adapter->hw.dev_spec._i225.eee_disable;
	error = sysctl_handle_int(oidp, &value, 0, req);
	if (error || req->newptr == NULL)
		return (error);

	adapter->hw.dev_spec._i225.eee_disable = (value != 0);
	igc_if_init(adapter->ctx);
#endif
	return (0);
}

static int
igc_sysctl_debug_info(SYSCTL_HANDLER_ARGS)
{
	//struct igc_adapter* adapter;
	int error = 0;
#if 0 //ToDo
	int result;
	result = -1;
	error = sysctl_handle_int(oidp, &result, 0, req);

	if (error || !req->newptr)
		return (error);

	if (result == 1) {
		adapter = (struct igc_adapter*)arg1;
		igc_print_debug_info(adapter);
	}
#endif
	return (error);
}

static int
igc_get_rs(SYSCTL_HANDLER_ARGS)
{
#if 0 //ToDo
	struct igc_adapter* adapter = (struct igc_adapter*)arg1;
	int error;
	int result;

	result = 0;
	error = sysctl_handle_int(oidp, &result, 0, req);

	if (error || !req->newptr || result != 1)
		return (error);
	igc_dump_rs(adapter);

	return (error);
#endif
	return 0;
}
#ifndef UNDER_RTSS
static void
igc_if_debug(if_ctx_t ctx)
{
	igc_dump_rs(iflib_get_softc(ctx));
}

#endif
/*
 * This routine is meant to be fluid, add whatever is
 * needed for debugging a problem.  -jfv
 */
static void
igc_print_debug_info(struct igc_adapter* adapter)
{
#if 0 //ToDo
	device_t dev = iflib_get_dev(adapter->ctx);
	struct ifnet* ifp = iflib_get_ifp(adapter->ctx);

	struct tx_ring* txr = &adapter->tx_queues->txr;
	struct rx_ring* rxr = &adapter->rx_queues->rxr;

	if (if_getdrvflags(ifp) & IFF_DRV_RUNNING)
		printf("Interface is RUNNING ");
	else
		printf("Interface is NOT RUNNING\n");

	if (if_getdrvflags(ifp) & IFF_DRV_OACTIVE)
		printf("and INACTIVE\n");
	else
		printf("and ACTIVE\n");

	for (int i = 0; i < adapter->tx_num_queues; i++, txr++) {
		device_printf(dev, "TX Queue %d ------\n", i);
		device_printf(dev, "hw tdh = %d, hw tdt = %d\n",
			IGC_READ_REG(&adapter->hw, IGC_TDH(i)),
			IGC_READ_REG(&adapter->hw, IGC_TDT(i)));

	}
	for (int j = 0; j < adapter->rx_num_queues; j++, rxr++) {
		device_printf(dev, "RX Queue %d ------\n", j);
		device_printf(dev, "hw rdh = %d, hw rdt = %d\n",
			IGC_READ_REG(&adapter->hw, IGC_RDH(j)),
			IGC_READ_REG(&adapter->hw, IGC_RDT(j)));
	}
#endif
}


static 
void* iflib_get_softc(if_ctx_t pCardInfo)
{
	return &pCardInfo->adapter;
}


void Rt_igc_if_init(void* pCardInfo)
{
	igc_if_stop(pCardInfo);

	igc_if_init(pCardInfo);
}

void Rt_igc_if_intr_enable(void* pCardInfo)
{
	igc_if_intr_enable(pCardInfo);
}

void Rt_igc_if_intr_disable(void* pCardInfo)
{
	igc_if_intr_disable(pCardInfo);
}

int Rt_igc_if_set_promisc(void *pCardInfo , int rtflags)
{
	return igc_if_set_promisc(pCardInfo, rtflags);
}

int RtCheckLink(void *pCardInfo, s32 *speed, s32 *link_up, s32 *duplex)
{
	struct igc_adapter* adapter = iflib_get_softc(pCardInfo);

	adapter->hw.mac.get_link_status = true;

	igc_if_update_admin_status(pCardInfo);
	
	*link_up = adapter->link_active;
	*duplex = adapter->link_duplex;
	*speed = adapter->link_speed;

	return IGC_SUCCESS;
}

int igc_print_stats_counters(struct igc_adapter* adapter)
{
	int i;
	RtPrintf("CRC Error Count adapter->stats.crcerrs %d \n", IGC_READ_REG(&adapter->hw, IGC_CRCERRS));
	RtPrintf("Alignment Error Count adapter->stats.algnerrc   %d \n", IGC_READ_REG(&adapter->hw, IGC_ALGNERRC));
	RtPrintf("RX Error Count   %d \n", IGC_READ_REG(&adapter->hw, IGC_RXERRC));
	RtPrintf("Missed packet count adapter->stats.mpc %d \n", IGC_READ_REG(&adapter->hw, IGC_MPC));

	RtPrintf("adapter->stats.scc %d \n", IGC_READ_REG(&adapter->hw, IGC_SCC));
	RtPrintf("adapter->stats.ecol %d \n", IGC_READ_REG(&adapter->hw, IGC_ECOL));

	RtPrintf("adapter->stats.mcc %d \n", IGC_READ_REG(&adapter->hw, IGC_MCC));
	RtPrintf("adapter->stats.latecol %d \n", IGC_READ_REG(&adapter->hw, IGC_LATECOL));
		RtPrintf("adapter->stats.colc %d \n", IGC_READ_REG(&adapter->hw, IGC_COLC));
		RtPrintf("adapter->stats.colc %d \n", IGC_READ_REG(&adapter->hw, IGC_RERC));
		RtPrintf("adapter->stats.dc  %d \n", IGC_READ_REG(&adapter->hw, IGC_DC));
		RtPrintf("adapter->stats.rlec  %d \n", IGC_READ_REG(&adapter->hw, IGC_RLEC));
		RtPrintf("adapter->stats.xonrxc %d \n", IGC_READ_REG(&adapter->hw, IGC_XONRXC));
		RtPrintf("adapter->stats.xoffrxc  %d \n", IGC_READ_REG(&adapter->hw, IGC_XOFFRXC));
		/*
		 * For watchdog management we need to know if we have been
		 * paused during the last interval, so capture that here.
		 */
		 //if (adapter->stats.xoffrxc != prev_xoffrxc)
		 //	adapter->shared->isc_pause_frames = 1;
		//RtPrintf("adapter->stats.xontxc  %d \n", IGC_READ_REG(&adapter->hw, IGC_XONTXC));
		//RtPrintf("adapter->stats.xofftxc  %d \n", IGC_READ_REG(&adapter->hw, IGC_XOFFTXC));
		RtPrintf("adapter->stats.fcruc  %d \n", IGC_READ_REG(&adapter->hw, IGC_FCRUC));
		RtPrintf("Packets Rx (64 bytes) adapter->stats.prc64  %d \n", IGC_READ_REG(&adapter->hw, IGC_PRC64));
		RtPrintf("Packets Rx (127 bytes) adapter->stats.prc127  %d \n", IGC_READ_REG(&adapter->hw, IGC_PRC127));
		RtPrintf("Packets Rx (255 bytes)adapter->stats.prc255  %d \n", IGC_READ_REG(&adapter->hw, IGC_PRC255));
		RtPrintf("Packets Rx (511 bytes)adapter->stats.prc511  %d \n", IGC_READ_REG(&adapter->hw, IGC_PRC511));
		RtPrintf("Packets Rx (1023 bytes)adapter->stats.prc1023  %d \n", IGC_READ_REG(&adapter->hw, IGC_PRC1023));
		RtPrintf("Packets Rx (1522 bytes) adapter->stats.prc1522  %d \n", IGC_READ_REG(&adapter->hw, IGC_PRC1522));
		RtPrintf("adapter->stats.tlpic  %d \n", IGC_READ_REG(&adapter->hw, IGC_TLPIC));
		RtPrintf("adapter->stats.rlpic  %d \n", IGC_READ_REG(&adapter->hw, IGC_RLPIC));
		RtPrintf("Good Packets Rx Count adapter->stats.gprc %d \n", IGC_READ_REG(&adapter->hw, IGC_GPRC));
		RtPrintf("Broadcast Packets Rx Count adapter->stats.bprc  %d \n", IGC_READ_REG(&adapter->hw, IGC_BPRC));
		RtPrintf("Multicast Packets Rx Count adapter->stats.mprc  %d \n", IGC_READ_REG(&adapter->hw, IGC_MPRC));
		RtPrintf("adapter->stats.gptc  %d \n", IGC_READ_REG(&adapter->hw, IGC_GPTC));

		/* For the 64-bit byte counters the low dword must be read first. */
		/* Both registers clear on the read of the high dword */

		RtPrintf("adapter->stats.gorc %d \n", IGC_READ_REG(&adapter->hw, IGC_GORCL) +
			((u64)IGC_READ_REG(&adapter->hw, IGC_GORCH) << 32));
		RtPrintf("adapter->stats.gotc  %d \n", IGC_READ_REG(&adapter->hw, IGC_GOTCL) +
			((u64)IGC_READ_REG(&adapter->hw, IGC_GOTCH) << 32));

		RtPrintf("adapter->stats.rnbc  %d \n", IGC_READ_REG(&adapter->hw, IGC_RNBC));
		RtPrintf("adapter->stats.ruc   %d \n", IGC_READ_REG(&adapter->hw, IGC_RUC));
		RtPrintf("adapter->stats.rfc   %d \n", IGC_READ_REG(&adapter->hw, IGC_RFC));
		RtPrintf("adapter->stats.roc   %d \n", IGC_READ_REG(&adapter->hw, IGC_ROC));
		RtPrintf("adapter->stats.rjc   %d \n", IGC_READ_REG(&adapter->hw, IGC_RJC));

		RtPrintf("adapter->stats.tor   %d \n", IGC_READ_REG(&adapter->hw, IGC_TORH));
		//RtPrintf("adapter->stats.tot   %d \n", IGC_READ_REG(&adapter->hw, IGC_TOTH));

		RtPrintf("adapter->stats.tpr   %d \n", IGC_READ_REG(&adapter->hw, IGC_TPR));

		for (i = 0; i < adapter->rx_num_queues; i++)
		{
			RtPrintf("ReceiveQueue (%d) IGC_PQGPRC  %d IGC_PQGORC %d IGC_PQMPRC %d IGC_RQDPC %d\n", i, 
							IGC_READ_REG(&adapter->hw, IGC_PQGPRC(i)),
							IGC_READ_REG(&adapter->hw, IGC_PQGORC(i)),
							IGC_READ_REG(&adapter->hw, IGC_PQMPRC(i)),
							IGC_READ_REG(&adapter->hw, IGC_RQDPC(i)));
		}

		RtPrintf("adapter->stats.tpt   %d \n", IGC_READ_REG(&adapter->hw, IGC_TPT));
		RtPrintf("adapter->stats.ptc64   %d \n", IGC_READ_REG(&adapter->hw, IGC_PTC64));
		RtPrintf("adapter->stats.ptc127   %d \n", IGC_READ_REG(&adapter->hw, IGC_PTC127));
		RtPrintf("adapter->stats.ptc255   %d \n", IGC_READ_REG(&adapter->hw, IGC_PTC255));
		RtPrintf("adapter->stats.ptc511   %d \n", IGC_READ_REG(&adapter->hw, IGC_PTC511));
		RtPrintf("adapter->stats.ptc1023   %d \n", IGC_READ_REG(&adapter->hw, IGC_PTC1023));
		RtPrintf("adapter->stats.ptc1522   %d \n", IGC_READ_REG(&adapter->hw, IGC_PTC1522));
		RtPrintf("adapter->stats.mptc   %d \n", IGC_READ_REG(&adapter->hw, IGC_MPTC));
		RtPrintf("adapter->stats.bptc   %d \n", IGC_READ_REG(&adapter->hw, IGC_BPTC));
		
		/* Interrupt Counts */
		RtPrintf("Interrupt Counts: \n");
		RtPrintf("adapter->stats.iac   %d \n", IGC_READ_REG(&adapter->hw, IGC_IAC));
		RtPrintf("adapter->stats.rxdmtc   %d \n", IGC_READ_REG(&adapter->hw, IGC_RXDMTC));
		RtPrintf("Mgmt Full Buffer Drop   %d \n", IGC_READ_REG(&adapter->hw, IGC_MANBUFDRP));
		RtPrintf("BMC2OS packets received by host   %d \n", IGC_READ_REG(&adapter->hw, IGC_B2OGPRC));
		RtPrintf("OS2BMC packets received by BMC   %d \n", IGC_READ_REG(&adapter->hw, IGC_O2BGPTC));
	RtPrintf("adapter->stats.tncrs   %d \n", IGC_READ_REG(&adapter->hw, IGC_TNCRS));
	RtPrintf("adapter->stats.htdpmc   %d \n", IGC_READ_REG(&adapter->hw, IGC_HTDPMC));
	RtPrintf("adapter->stats.tsctc   %d \n", IGC_READ_REG(&adapter->hw, IGC_TSCTC));

	return 0;
}